#!/usr/bin/env python

#cython: boundscheck=False
#cython: nonecheck=False
#cython: cdivision=True

"""
Cython module for Face definitions
"""

#maybe this is a cython .15 thing?
#from libc.math import INFINITY, M_PI, sqrt, pow, fabs, cos, sin, acos, atan2

cdef extern from "math.h":
    double M_PI
    double sqrt(double) nogil
    double atan2 (double y, double x )
    double pow(double x, double y)
    double fabs(double)
    double cos(double)
    double sin(double)
    double acos(double)

cdef extern from "float.h":
    double DBL_MAX

cdef double INF=(DBL_MAX+DBL_MAX)

from .ctracer cimport Face, sep_, \
        vector_t, ray_t, FaceList, subvv_, dotprod_, mag_sq_, norm_,\
            addvv_, multvs_, mag_, transform_t, Transform, transform_c,\
                rotate_c, Shape, Distortion

import numpy as np
cimport numpy as np_
cimport cython
from cython.parallel import prange

cdef struct flatvector_t:
    double x,y
    
    
cdef class ShapedFace(Face):
    cdef:
        public Shape shape
        public int invert_normals
        
    def __cinit__(self, **kwds):
        self.shape = kwds.get("shape", Shape())
        self.invert_normals = int(kwds.get('invert_normals', 0))
        
    cdef double eval_z_c(self, double x, double y) nogil:
        return 0.0
    
    cdef double eval_implicit_c(self, double x, double y, double z) nogil:
        return z - self.eval_z_c(x,y)
    
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    def eval_z_extent(self, double[:] x, double[:] y):
        """Samples the face surface z-values over the 2d grid given by the x- and y-
           arrays. Returns the max and min values of z"""
        cdef:
            size_t i,j,ni=x.shape[0], nj=y.shape[0]
            double maxv, minv, v
        maxv = self.eval_z_c(x[0],y[0])+0.1
        minv = maxv-0.2
        with nogil:
            for i in range(ni):
                for j in range(nj):
                    v = self.eval_z_c(x[i],y[j])
                    if v < minv:
                        minv = v
                    if v > maxv:
                        maxv = v
        return (minv, maxv)
    
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    def eval_z_points(self, double[:,:] points):
        cdef:
            size_t i, ni=points.shape[0]
            np_.ndarray aout = np.empty((ni,), dtype='d')
            double[:] out = aout
            
        with nogil:
            ###Not using prange here because ZernikeDistortion uses a cache for its calculation. 
            ###Threading screws this up.
            for i in prange(ni):
                out[i] = self.eval_z_c(points[i,0],points[i,1])
        return aout
    
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    def eval_implicit_grid(self, double[:] x, double[:] y, double[:] z):
        cdef:
            size_t i,j,k, ni=x.shape[0], nj=y.shape[0], nk=z.shape[0]
            np_.ndarray aout = np.empty((ni,nj,nk), dtype='d')
            double[:,:,:] out = aout
            int sign
            
        if self.invert_normals:
            sign = -1
        else:
            sign = 1
            
        with nogil:
            for i in prange(ni):
                for j in range(nj):
                    for k in range(nk):
                        out[i,j,k] = sign*self.eval_implicit_c(x[i], y[j], z[k])
        return aout
    
    @cython.boundscheck(False)  # Deactivate bounds checking
    @cython.wraparound(False)   # Deactivate negative indexing.
    def eval_implicit_points(self, double[:,:] points):
        cdef:
            size_t i, ni=points.shape[0]
            np_.ndarray aout = np.empty((ni,), dtype='d')
            double[:] out = aout
            int sign
        if self.invert_normals:
            sign = -1
        else:
            sign = 1
            
        with nogil:
            for i in prange(ni):
                out[i] = sign*self.eval_implicit_c(points[i,0],points[i,1],points[i,2])
        return aout
    

cdef class CircularFace(Face):
    cdef public double diameter, offset, z_plane
    
    params = ['diameter', 'offset']
    
    def __cinit__(self, **kwds):
        self.z_plane = kwds.get('z_plane', 0.0)
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          
        returns:
          the distance along the ray to the first valid intersection. No
          intersection can be indicated by a negative value.
        """
        cdef:
            double max_length = sep_(p1, p2)
            double h = (self.z_plane-p1.z)/(p2.z-p1.z)
            double X, Y, d=self.diameter
            
        #print "CFACE", p1, p2
        
        if (h<self.tolerance) or (h>1.0):
            #print "H", h
            return 0
        X = p1.x + h*(p2.x-p1.x) - self.offset
        Y = p1.y + h*(p2.y-p1.y)
        if is_base_ray and (X*X + Y*Y) > (d*d/4):
            #print "X", X, "Y", Y
            return 0
        return h * max_length

    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef vector_t normal
        normal.x=0
        normal.y=0
        normal.z=-1
        return normal
    
    
cdef class ShapedPlanarFace(ShapedFace):
    cdef:
        public double z_height
        
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          
        returns:
          the distance along the ray to the first valid intersection. No
          intersection can be indicated by a negative value.
        """
        cdef:
            double max_length = sep_(p1, p2)
            double h = (self.z_height-p1.z)/(p2.z-p1.z)
            double X, Y
                    
        if (h<self.tolerance) or (h>1.0):
            return -1
        
        X = p1.x + h*(p2.x-p1.x)
        Y = p1.y + h*(p2.y-p1.y)
        if is_base_ray and not (<Shape>self.shape).point_inside_c(X,Y):
            return -1
        return h*max_length
        
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef vector_t normal
        normal.x=0
        normal.y=0
        normal.z=1
        return normal
    
    cdef double eval_z_c(self, double x, double y) nogil:
        return self.z_height
    
    cdef double eval_implicit_c(self, double x, double y, double z) nogil:
        return (z - self.z_height)
    
    
cdef class ElipticalPlaneFace(Face):
    cdef public double g_x, g_y, diameter
    
    params = ['diameter']
    
    def __cinit__(self, **kwds):
        self.g_x = kwds.get('g_x', 0.0)
        self.g_y = kwds.get('g_y', 0.0)
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double max_length = sep_(p1, p2)
            double h = (self.g_x*p1.x + self.g_y*p1.y - p1.z) / \
                        ((p2.z-p1.z) - self.g_x*(p2.x-p1.x) - self.g_y*(p2.y-p1.y))
            double X,Y,Z, d=self.diameter
            
        if (h<self.tolerance) or (h>1.0):
            return 0
        
        X = p1.x + h*(p2.x-p1.x)
        Y = p1.y + h*(p2.y-p1.y)
        Z = p1.z + h*(p2.z-p1.z)
        if is_base_ray and (X*X + Y*Y) > (d*d/4):
            #print "X", X, "Y", Y
            return 0
        return h * max_length
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef vector_t normal
        normal.x=self.g_x
        normal.y=self.g_y
        normal.z=-1
        return norm_(normal)
    
    
cdef class RectangularFace(Face):
    cdef public double length, width, offset, z_plane
    
    params = ['length', 'width', 'offset']
    
    def __cinit__(self, **kwds):
        self.z_plane = kwds.get('z_plane', 0.0)
        self.width = kwds.get("width", 2.0)
        self.length = kwds.get("length", 5.0)
        self.offset = kwds.get("offset", 0.0)
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          ray - pointer to the input ray
          
        returns:
          idx - -1 if no intersection is found *or* the distance to the 
                intersection is larger than the existing ray.length. OTherwise,
                this is set to the intersecting face idx
        """
        cdef:
            double max_length = sep_(p1, p2)
            double h = (self.z_plane-p1.z)/(p2.z-p1.z)
            double X, Y, lngth=self.length, wdth = self.width
            
        if (h<self.tolerance) or (h>1.0):
            return 0
        
        if is_base_ray:
            X = p1.x + h*(p2.x-p1.x) - self.offset
            Y = p1.y + h*(p2.y-p1.y)
            
            #if x or y displacement is greater than length or width of rectangle, no intersect
            if X*X > lngth*lngth/4:
                return 0
            if Y*Y > wdth*wdth/4:
                return 0 
        return h * max_length

    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef vector_t normal
        normal.x=0
        normal.y=0
        normal.z=-1
        return normal
    
    
cdef class SphericalFace(Face):
    cdef public double diameter, curvature, z_height
    
    #Don't want curvature in this list in case the owner defines the 
    #curvature in a different way
    params = ['diameter',] 
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
    
    cdef double intersect_c(self, vector_t r, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          r - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          ray - pointer to the input ray
          
        returns:
          idx - -1 if no intersection is found *or* the distance to the 
                intersection is larger than the existing ray.length. OTherwise,
                this is set to the intersecting face idx
        """
        cdef:
            double A,B,C,D, cz, a1, a2
            vector_t s, d, pt1, pt2
            
        s = subvv_(p2, r)
        cz = self.z_height - self.curvature
        d = r
        d.z -= cz
        
        A = mag_sq_(s)
        B = 2*dotprod_(s,d)
        C = mag_sq_(d) - self.curvature**2
        D = B*B - 4*A*C
        
        if D < 0: #no intersection with sphere
            return 0
        
        D = sqrt(D)
        
        #1st root
        a1 = (-B+D)/(2*A) 
        pt1 = addvv_(r, multvs_(s, a1))
        #2nd root
        a2 = (-B-D)/(2*A)
        pt2 = addvv_(r, multvs_(s, a2))
        
        if self.curvature >= 0:
            if pt1.z < cz:
                a1 = INF
            if pt2.z < cz:
                a2 = INF
        else:
            if pt1.z > cz:
                a1 = INF
            if pt2.z > cz:
                a2 = INF
            
        D = self.diameter*self.diameter/4.
        
        if is_base_ray:
            if (pt1.x*pt1.x + pt1.y*pt1.y) > D:
                a1 = INF
            if (pt2.x*pt2.x + pt2.y*pt2.y) > D:
                a2 = INF
        
        if a2 < a1:
            a1 = a2
        
        if a1>1.0 or a1<self.tolerance:
            return 0
        return a1 * sep_(r, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        
        p.z -= (self.z_height - self.curvature)
        if self.curvature < 0:
            p.z = -p.z
            p.y = -p.y
            p.x = -p.x
        return norm_(p)
    
    
cdef class ShapedSphericalFace(ShapedFace):
    cdef:
        public double curvature, z_height
    
    #Don't want curvature in this list in case the owner defines the 
    #curvature in a different way
    params = [] 
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
        self.curvature = kwds.get("curvature", 100.0)
    
    cdef double intersect_c(self, vector_t r, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          r - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          ray - pointer to the input ray
          
        returns:
          idx - -1 if no intersection is found *or* the distance to the 
                intersection is larger than the existing ray.length. OTherwise,
                this is set to the intersecting face idx
        """
        cdef:
            double A,B,C,D, cz, a1, a2
            vector_t s, d, pt1, pt2
            
        s = subvv_(p2, r)
        cz = self.z_height - self.curvature
        d = r
        d.z -= cz
        
        A = mag_sq_(s)
        B = 2*dotprod_(s,d)
        C = mag_sq_(d) - self.curvature**2
        D = B*B - 4*A*C
        
        if D < 0: #no intersection with sphere
            return 0
        
        D = sqrt(D)
        
        #1st root
        a1 = (-B+D)/(2*A) 
        pt1 = addvv_(r, multvs_(s, a1))
        #2nd root
        a2 = (-B-D)/(2*A)
        pt2 = addvv_(r, multvs_(s, a2))
        
        if self.curvature >= 0:
            if pt1.z < cz:
                a1 = INF
            if pt2.z < cz:
                a2 = INF
        else:
            if pt1.z > cz:
                a1 = INF
            if pt2.z > cz:
                a2 = INF
           
        if is_base_ray:
            if not (<Shape>self.shape).point_inside_c( pt1.x, pt1.y ):
                a1 = INF
            if not (<Shape>self.shape).point_inside_c( pt2.x, pt2.y ):
                a2 = INF
        
        if a2 < a1:
            a1 = a2
        
        if a1>1.0 or a1<self.tolerance:
            return 0
        return a1 * sep_(r, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        
        p.z -= (self.z_height - self.curvature)
        if self.curvature < 0:
            p.z = -p.z
            p.y = -p.y
            p.x = -p.x
        return norm_(p)
    
    cdef double eval_z_c(self, double x, double y) nogil:
        cdef:
            double r2 = x*x + y*y
            double c = self.curvature
        if c >= 0:
            return self.z_height + sqrt(c*c - r2) - c
        else:
            return self.z_height - sqrt(c*c - r2) - c
    
    cdef double eval_implicit_c(self, double x, double y, double z) nogil:
        cdef:
            double cz, c=self.curvature

        cz = z - self.z_height + c
        
        if c >= 0:
            return ((x*x + y*y + cz*cz) - (c*c))
        else:
            return -((x*x + y*y + cz*cz) - (c*c))
    
    
cdef class ExtrudedPlanarFace(Face):
    cdef:
        double x1_, y1_, x2_, y2_
        public double z1, z2
        vector_t normal
        
    def __cinit__(self, **kwds):
        self.x1 = kwds.get('x1',0)
        self.y1 = kwds.get('y1',0)
        self.x2 = kwds.get('x2',0)
        self.y2 = kwds.get('y2',0)
        self.z1 = kwds.get('z1',0)
        self.z2 = kwds.get('z2',0)
        
    property x1:
        def __get__(self):
            return self.x1_
        
        def __set__(self, double v):
            self.x1_ = v
            self.calc_normal()
            
    property y1:
        def __get__(self):
            return self.y1_
        
        def __set__(self, double v):
            self.y1_ = v
            self.calc_normal()
            
    property x2:
        def __get__(self):
            return self.x2_
        
        def __set__(self, double v):
            self.x2_ = v
            self.calc_normal()
            
    property y2:
        def __get__(self):
            return self.y2_
        
        def __set__(self, double v):
            self.y2_ = v
            self.calc_normal()
            
    cdef calc_normal(self):
        cdef vector_t n
            
        n.y = self.x2_ - self.x1_
        n.x = self.y1_ - self.y2_
        n.z = 0
        self.normal = norm_(n)
        
    cdef double intersect_c(self, vector_t r, vector_t p2, int is_base_ray):
        cdef: 
            vector_t s, u, v
            double a, dz
            
        u.x = self.x1_
        u.y = self.y1_
        
        v.x = self.x2_ - u.x
        v.y = self.y2_ - u.y
        
        s = subvv_(p2, r)
        
        if is_base_ray:
            #fractional distance of intersection along edge
            a = (s.y*(u.x-r.x) - s.x*(u.y-r.y)) / (s.x*v.y - s.y*v.x)
            
            if a<0:
                return 0
            if a>1:
                return 0
        #distance of intersection along ray (in XY plane)
        a = (v.x*(r.y-u.y) - v.y*(r.x-u.x)) / (s.x*v.y - s.y*v.x)
        
        if is_base_ray:
            #distance in 3D
            dz = a*(p2.z - r.z)
            if self.z1 < (r.z+dz) < self.z2:
                return a * mag_(s)
            else:
                return 0
        else:
            return a * mag_(s)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        return self.normal

#
# Functions used for Bezier math.  should go in utils.pyx when there is one
#


    
cdef double eval_bezier(double t, double cp0, double cp1, double cp2, double cp3):
    #just evaluate a cubic bezier spline
    return cp0*((1-t)**3) + 3*cp1*t*((1-t)**2) + 3*cp2*(1-t)*(t**2) + cp3*(t**3)

cdef double dif_bezier(double t, double cp0, double cp1, double cp2, double cp3):
    #calc the derivative of a cubic bezier when parameter = t
    cdef long double A, B, C     #just doin this old school polynomial style
    A = cp3-3*cp2+3*cp1-cp0
    B = 3*cp2-6*cp1+3*cp0
    C = 3*cp1-3*cp0
    return 3*A*t**2 + 2*B*t + C

cdef struct poly_roots:
    #my apologies for highly unusable code, 
    #I am such a noob at passing values between functions
    double roots[3] 
    int n
    
cdef poly_roots roots_of_cubic(double a, double b, double c, double d):
    #this code is known not to work in the case of (x-c)^3 (triple zero)
    # **TODO ** fix this
    # TODO: cubic solution explodes with small a, co quadratic is used.
    # using newton's root finding method could quickly polish up answer.
    cdef:
        long double a1 = b/a, a2 = c/a, a3 = d/a
        long double Q = (a1*a1 - 3.0*a2)/9.0
        long double R = (2.0*a1*a1*a1 - 9.0*a1*a2 + 27.0*a3)/54.0
        long double R2_Q3 = R*R - Q*Q*Q
        long double theta
        poly_roots x= ((0.0,0.0,0.0),0)
    if fabs(a) <= 0.0000000001:
        #^this, precision is less than ideal here
        if fabs(b) <= 0.0000000001:
            if c == 0:
                #our "polynomial" is a constant
                # return a nonsense answer
                x.n = 1
                x.roots[0] = 0
            else:
                #print "solved for line"
                #our polynomial is a line cx+d
                x.n = 1
                x.roots[0] = -d/c
        else:
            #print "solved for quadratic"
            #our ploynomial is quadratic, bx**2+cx+d
            a1 = c**2 - 4*b*d
            a1 = sqrt(a1)
            x.n = 2
            x.roots[0] = (-c+a1)/(2*b)
            x.roots[1] = (-c-a1)/(2*b)
            #print "roots: ",x.roots[0],x.roots[1]
    else:
        #our polynomial is actually cubic   
        #original code had <= here, but if determinant is 0 then only one answer
        if (R2_Q3 < 0):
            x.n = 3
            theta = acos(R/sqrt(Q*Q*Q))
            x.roots[0] = -2.0*sqrt(Q)*cos(theta/3.0) - a1/3.0;
            x.roots[1] = -2.0*sqrt(Q)*cos((theta+2.0*M_PI)/3.0) - a1/3.0
            x.roots[2] = -2.0*sqrt(Q)*cos((theta+4.0*M_PI)/3.0) - a1/3.0
            #print "triple",x.roots[0],x.roots[1],x.roots[2]
        else:
            x.n = 1
            a2 = pow(sqrt(R2_Q3)+fabs(R), 1/3.0)
            a2 += Q/x.roots[0]
            a2 *= (1 if (R < 0.0) else -1)
            a2 -= a1/3.0
            x.roots[0] = a2 
            #print "single: ",x.roots[0]
    return x

cdef flatvector_t rotate2D(double phi, flatvector_t p):
    cdef flatvector_t result
    result.x = p.x*cos(phi) - p.y*sin(phi)
    result.y = p.x*sin(phi) + p.y*cos(phi)     
    return result

cdef class ExtrudedBezierFace(Face):

    cdef:
        double z_height_1, z_height_2
        flatvector_t mincorner, maxcorner         #corners of x-y box that bounds entire spline
        np_.ndarray curves_array        

    cdef int ccw(self, flatvector_t  A, flatvector_t  B, flatvector_t  C):
        #used by an ingenious line segment intersect algorithim I found.
        #determines counter clockwiseness of points
        return (C.y-A.y)*(B.x-A.x) > (B.y-A.y)*(C.x-A.x)

    cdef int line_seg_overlap(self, flatvector_t A, flatvector_t B, flatvector_t C, flatvector_t D):
        #check if two line segments overlap eachother.  Used for rough tests of
        #intersection before committing to much computation to the potential intersection.
        # A and B are the begin and end of one line; C,D the other. 
        # Code from http://www.bryceboe.com/2006/10/23/line-segment-intersection-algorithm/
        #AB crosses CD if ABCD are all cw or ccw.
        return self.ccw(A,C,D) != self.ccw(B,C,D) and self.ccw(A,B,C) != self.ccw(A,B,D)
        
    cdef int pnt_in_hull(self,flatvector_t p, flatvector_t A, flatvector_t B, flatvector_t C, flatvector_t D):
        #instead of convex hull, just use xy bounding box, which is quicker to construct

        cdef int i,j,k
        cdef float w

        #if p is bigger than atleast one but not all, the it is in the box
        i = p.x > A.x or p.x>B.x or p.x>C.x or p.x>D.x
        j = p.x > A.x and p.x>B.x and p.x>C.x and p.x>D.x
        k = i and not j
        i = p.y > A.y or p.y>B.y or p.y>C.y or p.y>D.y
        j = p.y > A.y and p.y>B.y and p.y>C.y and p.y>D.y
        i = i and not j

        #maybe a bad idea, but this is incase bezier is actually a plane in x or y only:
        w = A.x - D.x
        w = w*w
        if w <= .0005:
            w = B.x - A.x
            w=w*w
            if w <= .0005:
                i=k=True
        else:
            w = A.y - D.y
            w = w*w
            if w <= .0005:
                w = B.y - A.y
                w=w*w
                if w <= .0005:
                    i=k=True
        return i and k

    def __cinit__(self,np_.ndarray[np_.float64_t ,ndim=3] beziercurves,double z_height_1=0,double z_height_2=0, **kwds):
        cdef flatvector_t temp1, temp2
        self.curves_array = beziercurves
        self.z_height_1 = z_height_1
        self.z_height_2 = z_height_2
        temp1.x = beziercurves[0,0,0]
        temp1.y = beziercurves[0,0,1]
        temp2.x = beziercurves[0,0,0]
        temp2.y = beziercurves[0,0,1]

        for bezierpts in beziercurves:
            for pair in bezierpts:
                if pair[0] < temp1.x: temp1.x=pair[0]
                if pair[0] > temp2.x: temp2.x=pair[0]
                if pair[1] < temp1.y: temp1.y=pair[1]
                if pair[1] > temp2.y: temp2.y=pair[1]

        self.mincorner = temp1
        self.maxcorner = temp2

    cdef double intersect_c(self, vector_t ar, vector_t pee2, int is_base_ray):

        cdef: 
            flatvector_t tempvector
            flatvector_t r, p2, s, origin
            flatvector_t cp0,cp1,cp2,cp3  #holds control points for spline segment under scrutiny
            double result = INF            #length of ray before it intersects surface. 0 if no valid intersection
            double dZ                       #rate of change of z. dZ*result+Z0 gives Z coordinate
            double A,B,C,D,t,a,b,c,d
            poly_roots ts
            vector_t    tempv
        #print "\ncalled intersection"

        origin.x = 0 
        origin.y = 0
        #first off, does ray even enter the depth of the extrusion?
        if (ar.z < self.z_height_1 and pee2.z <self.z_height_1) or (ar.z > self.z_height_2 and pee2.z > self.z_height_2):
            return 0    #does not
        #print "ray passes z test"
        #strip useless thrid dimension from ray vector
        r.x = ar.x
        r.y = ar.y
        p2.x = pee2.x
        p2.y = pee2.y


        #check if ray intersects x-y bounding box of spline at all
        tempvector.x = self.mincorner.x
        tempvector.y = self.maxcorner.y
        if not self.line_seg_overlap(r,p2,self.mincorner,tempvector):
            if not  self.line_seg_overlap(r,p2,tempvector,self.maxcorner):
                tempvector.x = self.maxcorner.x
                tempvector.y = self.mincorner.y
                if not self.line_seg_overlap(r,p2,self.maxcorner,tempvector):
                    if not self.line_seg_overlap(r,p2,tempvector,self.mincorner):
                        return 0    #no intersections
        #print "ray is in big bounding box"
        #segment intersects with gross bounding box,
        #Calc dZ and the 2D origin adjusted ray, because they will probably be used.
        tempv = subvv_ (pee2,ar) 
        dZ = tempv.z
        s.x=tempv.x
        s.y=tempv.y
        theta = atan2(s.y,s.x)
                
        s = rotate2D(-theta,s)
        # now, loop through curves and see if segment 1) intersects with individual convex hulls
        # 2) intersects with spline (return points)
        for curve in self.curves_array.copy():            #load up control points
            for pt in curve:
                pt[0]=pt[0]-ar.x
                pt[1]=pt[1]-ar.y 
            cp0.x,cp0.y = curve[0].copy()
            cp1.x,cp1.y = curve[1].copy()
            cp2.x,cp2.y = curve[2].copy()
            cp3.x,cp3.y = curve[3].copy()
            
            
            #rotate ctrl points such that ray is along the x axis
            cp0 = rotate2D(-theta,cp0)
            cp1 = rotate2D(-theta,cp1)
            cp2 = rotate2D(-theta,cp2)
            cp3 = rotate2D(-theta,cp3)
            #test for intersection between ray (actually segment) and convex hull
            if self.line_seg_overlap(origin,s,cp0,cp1) or self.line_seg_overlap(origin,s,cp1,cp2) or self.line_seg_overlap(origin,s,cp2,cp3) or self.line_seg_overlap(origin,s,cp3,cp0):
                #print "inside intersect hull"
                #Ray does intersect this convex hull.  Find solution:
                #Setup A,B,C and D (bernstein polynomials)
                A = cp3.y-3*cp2.y+3*cp1.y-cp0.y
                B = 3*cp2.y-6*cp1.y+3*cp0.y
                C = 3*cp1.y-3*cp0.y
                D = cp0.y
                #solve for t
                #print "ABCD: ",A,B,C,D
                ts = roots_of_cubic(A,B,C,D)
                while ts.n > 0:
                    ts.n-=1
                    t = ts.roots[ts.n]
                    #print "root: ", t
                    #make sure solution is on valid interval
                    if 0.<t<1.:
                        #the x value will also be the length, which is the form of result
                        b = eval_bezier(t,cp0.x,cp1.x,cp2.x,cp3.x)
                        #print "b at t",b,t
                        #is x within bounds?
                        if 0 < b < s.x:
                            #print "in range"
                            #is point within Z bounds?
                            c = dZ*b/s.x   
                            a = c+ar.z     
                            if self.z_height_1 < a < self.z_height_2:
                                #print "in z: ",B,result
                                #is this the shortest length to an intersection so far?
                                b = sqrt(c**2+b**2)
                                #print "this: ",b
                                if b < result and b > self.tolerance:
                                    result = b
                                    #print "b at t",b,t

        if result == INF: result = 0

        return result



    cdef vector_t compute_normal_c(self, vector_t p):
        cdef:
            flatvector_t ray,cp0,cp1,cp2,cp3,rotated
            double theta, tmp, t
            poly_roots ts
            np_.ndarray box
        #print "called looking for normal",p.x,p.y
        ray.x = p.x
        ray.y = p.y
        theta = atan2(p.y,p.x)
        #find which curve this point is in
        for curve in self.curves_array.copy():            #load up control points
            cp0.x,cp0.y = curve[0].copy()
            cp1.x,cp1.y = curve[1].copy()
            cp2.x,cp2.y = curve[2].copy()
            cp3.x,cp3.y = curve[3].copy()

            #is point even in this hull?
            #print "tried at least"
            if self.pnt_in_hull(ray,cp0,cp1,cp2,cp3):
                #then, solve for t
                #print "in hull"
                cp0 = rotate2D(-theta,cp0)
                cp1 = rotate2D(-theta,cp1)
                cp2 = rotate2D(-theta,cp2)
                cp3 = rotate2D(-theta,cp3)
                
                #Setup A,B,C and D (bernstein polynomials)
                A = cp3.y-3*cp2.y+3*cp1.y-cp0.y
                B = 3*cp2.y-6*cp1.y+3*cp0.y
                C = 3*cp1.y-3*cp0.y
                D = cp0.y
                ts = roots_of_cubic(A,B,C,D)
                
                #print "normal roots: ",ts.n,ts.roots[0],ts.roots[1],ts.roots[2]
                while ts.n > 0:
                    ts.n -=1
                    t = ts.roots[ts.n]
                    #make sure solution is within interval
                    #print "normal t: ",t
                    if 0<=t<=1:
                        #ok, then is this the t to the same point p? 
                        tmp = eval_bezier(t,cp0.x,cp1.x,cp2.x,cp3.x)
                        #print "normal b at t: ",tmp,t
                        #I will generously allow for rounding error 
                        #print "got here with point: ",tmp,
                        if tmp**2 - (ray.x**2+ray.y**2) < .0001:
                            #this is the single solution. return the derivative dy/dx = dy/dt / dx/dt
                            #print "that was it!"
                            ray.x = dif_bezier(t,cp0.x,cp1.x,cp2.x,cp3.x)
                            ray.y = dif_bezier(t,cp0.y,cp1.y,cp2.y,cp3.y)
                            ray = rotate2D(theta,ray)
                            p.z = 0     #trough has no slope in z
                            #direction of normal is to the left of the parametric curve
                            #slope of normal is -dx/dy
                            
                            if ray.y==0:
                                p.x=0
                                p.y = (1 if ray.x>0 else -1)
                            elif ray.y>0:
                                p.x=-1
                                p.y = ray.x/ray.y
                            elif ray.y<0:
                                p.x = 1
                                p.y = -ray.x/ray.y    
                            return norm_(p)


        #how did you get here?  p was supposed to be a point on the curve!
        print("error: Bezier normal not found, point not actually on curve!")
        p.x=p.y=p.z = 0
        return p


    
cdef int point_in_polygon_c(double X, double Y, object obj):
    cdef int i, size, ct=0
    cdef double y1, y2, h, x, x1, x2
    cdef np_.ndarray[np_.float64_t, ndim=2] pts=obj
    
    size = pts.shape[0]
    
    y1 = pts[size-1,1]
    x1 = pts[size-1,0]
    for i in xrange(size):
        y2 = pts[i,1]
        x2 = pts[i,0]
        h = (Y - y1) / (y2 - y1)
        if 0 < h <= 1.0:
            x = x1 + h*(x2 - x1)
            if x > X:
                ct = not ct
        y1 = y2
        x1 = x2
    return ct


def point_in_polygon(double X, double Y, point_list):
    pts = np.ascontiguousarray(point_list, dtype=np.float64)
    assert pts.shape[1]==2
    return bool(point_in_polygon_c(X, Y, pts))
    
    
cdef class PolygonFace(Face):
    cdef public double z_plane
    cdef object _xy_points
    
    def __cinit__(self, z_plane=0.0, xy_points=[[]], **kwds):
        self.z_plane = z_plane
        self.xy_points = xy_points
    
    property xy_points:
        def __get__(self):
            return self._xy_points
        
        def __set__(self, pts):
            data = np.ascontiguousarray(pts, dtype=np.float64).reshape(-1,2)
            self._xy_points=data
            
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double max_length = sep_(p1, p2)
            double h = (self.z_plane-p1.z)/(p2.z-p1.z)
            double X, Y
        
        if (h<self.tolerance) or (h>1.0):
            #print "H", h
            return 0
        X = p1.x + h*(p2.x-p1.x)
        Y = p1.y + h*(p2.y-p1.y)
        #test for (X,Y) in polygon
        if is_base_ray and point_in_polygon_c(X,Y, self._xy_points)==1:
            return h * max_length
        else:
            return 0.0
        
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef vector_t normal
        normal.x=0
        normal.y=0
        normal.z=-1
        return normal
    
    
cdef class OffAxisParabolicFace(Face):
    cdef:
        public double EFL, diameter, height
                
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          
        returns:
          the distance along the ray to the first valid intersection. No
          intersection can be indicated by a negative value.
        """
        cdef:
            double max_length = sep_(p1, p2)
            double A = 1 / (2*self.EFL), efl = self.EFL
            double a,b,c,d
            vector_t s, r, pt1, pt2
            
            
        s = subvv_(p2,p1)
        r = p1
        r.z += efl/2.
                
        a = A*(s.x**2 + s.y**2)
        b = 2*A*(r.x*s.x + r.y*s.y) - s.z
        c = A*(r.x**2 + r.y**2) - r.z
        d = b**2 - 4*a*c
        
        ###FIXME
        if d<0: #no intersection ###BROKEN
            return 0
        
        if a < 1e-10: #approximate to zero if we're close to the parabolic axis
            a1 = -c/b
            pt1 = addvv_(r, multvs_(s, a1))
            pt1.x -= efl
            if (pt1.x*pt1.x + pt1.y*pt1.y) > (self.diameter/2):
                return 0
            if a1>1.0 or a1<self.tolerance:
                return 0
            return a1 * sep_(p1, p2)
        
        else:
            d = sqrt(d)
           
            #1st root
            a1 = (-b+d)/(2*a) 
            pt1 = addvv_(r, multvs_(s, a1))
            #2nd root
            a2 = (-b-d)/(2*a)
            pt2 = addvv_(r, multvs_(s, a2))
                           
            pt1.x -= efl
            pt2.x -= efl
            
            if is_base_ray:
                d = self.diameter
                d *= d/4.
                
                if (pt1.x*pt1.x + pt1.y*pt1.y) > d:
                    a1 = INF
                if (pt2.x*pt2.x + pt2.y*pt2.y) > d:
                    a2 = INF
            
            if a2 < a1:
                a1 = a2
            
            if a1>1.0 or a1<self.tolerance:
                return 0
            return a1 * sep_(p1, p2)
#
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef:
            vector_t normal
            double A = 1 / (2*self.EFL)
            double B, dz, m2
        
        m2 = p.x*p.x + p.y*p.y
        B = 4*m2*A*A
        dz = -sqrt( B/(B+1) )
        m2 = sqrt(m2)
        
        normal.x= -(dz * p.x)/m2
        normal.y= -(dz * p.y)/m2
        normal.z= -1 / sqrt( B+1 )
        return normal
    
        
cdef class EllipsoidalFace(Face):
    cdef:
        public double major, minor #axis lengths
        transform_t trans, inv_trans
        public double x1, x2, y1, y2, z1, z2 #local bounds of the ellpsoid block
        
    property transform:
        def __get__(self):
            t = Transform()
            t.trans = self.trans
            return t
        
        def __set__(self, Transform t):
            self.trans = t.trans
            
    property inverse_transform:
        def __set__(self, Transform t):
            self.inv_trans = t.trans
           
        def __get__(self):
            cdef Transform t=Transform()
            t.trans = self.inv_trans
            return t
            
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double B,A, a, b, c, d
            
            vector_t S = subvv_(p2, p1)
            vector_t r = transform_c(self.trans, p1)
            vector_t s = transform_c(self.trans, p2)
            
        s = subvv_(s, r)
        
        B = self.minor**2
        A = self.major**2
        
        a = A*(s.z*s.z + s.y*s.y) + B*s.x*s.x
        b = 2*( A*(r.z*s.z + r.y*s.y) + B*r.x*s.x )
        c = A*(r.z*r.z + r.y*r.y) + B*r.x*r.x - A*B
        
        d = b*b - 4*a*c
        d = sqrt(d)
        root1 = (-b + d)/(2*a)
        root2 = (-b - d)/(2*a)
        p2 = addvv_(p1, multvs_(S, root2))
        p1 = addvv_(p1, multvs_(S, root1))
        
        if is_base_ray:
            if not self.x1 < p2.x < self.x2:
                root2 = 2
            if not self.y1 < p2.y < self.y2:
                root2 = 2
            if not self.z1 < p2.z < self.z2:
                root2 = 2
                
            if not self.x1 < p1.x < self.x2:
                root1 = 2
            if not self.y1 < p1.y < self.y2:
                root1 = 2
            if not self.z1 < p1.z < self.z2:
                root1 = 2
        
        if root1 < self.tolerance:
            root1 = 2
        if root2 < self.tolerance:
            root2 = 2
        if root1 > root2:
            root1 = root2
        if root1 > 1:
            return 0
        return root1*mag_(S)
        
    cdef vector_t compute_normal_c(self, vector_t p):
        cdef vector_t n
        
        p = transform_c(self.trans, p)
        
        n.x = p.x/-(self.major**2)
        n.y = p.y/-(self.minor**2)
        n.z = p.z/-(self.minor**2)
        
        n = rotate_c(self.inv_trans, n)
        return norm_(n)
    
    def update(self):
        super(EllipsoidalFace, self).update()
        owner = self.owner
        self.sync_transform(owner.ellipse_trans)
        self.major, self.minor = owner.axes
        self.x1, self.x2 = owner.X_bounds
        self.y1, self.y2 = owner.Y_bounds
        self.z1, self.z2 = owner.Z_bounds
        
    def sync_transform(self, vtk_trans):
        m = vtk_trans.matrix
        rot = [[m.get_element(i,j) for j in xrange(3)] for i in xrange(3)]
        dt = [m.get_element(i,3) for i in xrange(3)]
        #print "TRANS", rot, dt
        self.transform = Transform(rotation=rot, translation=dt)
        inv_trans = vtk_trans.linear_inverse
        m = inv_trans.matrix
        rot = [[m.get_element(i,j) for j in xrange(3)] for i in xrange(3)]
        dt = [m.get_element(i,3) for i in xrange(3)]
        self.inverse_transform = Transform(rotation=rot, translation=dt)
        
        
cdef class SaddleFace(ShapedFace):
    cdef:
        public double z_height, curvature
        
    params=[]
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get("z_height", 0.0)
        self.curvature = kwds.get("curvature", 0.0)
        
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double A=sqrt(6.0), root, denom, a1, a2
            vector_t d,p, pt1, pt2
            
        A *= self.curvature
        p = p1
        p.z -= self.z_height
        d = subvv_(p2,p1)
        
        if d.x==0.0:
            a1 = (-A*(p.x*p.y) + p.z)/(A*d.y*p.x - d.z)
            a2 = INF
        elif d.y==0.0:
            a1 = (-A*(p.x*p.y) + p.z)/(A*d.x*p.y - d.z)
            a2 = INF
        else:
            root = (A**2)*(d.x**2)*(p.y**2) - 2*(A**2)*d.x*d.y*p.x*p.y + (A**2)*(d.y**2)*(p.x**2) + \
                    4*A*d.x*d.y*p.z - 2*A*d.x*d.z*p.y - 2*A*d.y*d.z*p.x + d.z**2
                    
            if root < 0:
                return -1
            
            root = sqrt(root)
            denom = 2*A*(d.x*d.y)
            
            a1 = a2 = -A*d.x*p.y - A*d.y*p.x + d.z
            a1 += root
            a2 -= root
            a1 /= denom
            a2 /= denom
        
        #print("a1:", a1, "a2:", a2, "root:", root, "sep:", sep_(p1,p2))
        
        pt1 = addvv_(p1, multvs_(d, a1))
        pt2 = addvv_(p1, multvs_(d, a2))
        
        if a1 < 0.0:
            a1 = INF
        if a2 < 0.0:
            a2 = INF
        
        if is_base_ray:
            if not (<Shape>self.shape).point_inside_c( pt1.x, pt1.y ):
                a1 = INF
            if not (<Shape>self.shape).point_inside_c( pt2.x, pt2.y ):
                a2 = INF
        
        if a2 < a1:
            a1 = a2
        
        if a1>1.0 or a1<self.tolerance:
            return -1
        return a1 * sep_(p1, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef:
            vector_t n
            double rt6 = sqrt(6)*self.curvature
        
        n.z = 1.0
        n.x = -rt6*p.y
        n.y = -rt6*p.x
        return norm_(n)
    
    cdef double eval_z_c(self, double x, double y) nogil:
        return sqrt(6) * self.curvature * x * y
        
        
    
        
cdef class CylindericalFace(ShapedFace):
    cdef:
        public double z_height, radius
    
    params = [] 
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
        self.radius = kwds.get("radius", 100.0)
        
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double a1, a2, cz, ox2, oz2, dx2, dz2, denom, R=self.radius
            double R2 = R*R
            vector_t d,o, pt1, pt2
            
        o = p1
        o.z -= self.z_height
        d = subvv_(p2,p1)
        
        ox2 = o.x*o.x
        oz2 = o.z*o.z
        dx2 = d.x*d.x
        dz2 = d.z*d.z
        
        root = R2*dz2 - 2*R*dx2*o.z + 2*R*d.x*d.z*o.x - dx2*oz2 + 2*d.x*d.z*o.x*o.z - dz2*ox2
        
        if root < 0:
            return -1
        
        root = sqrt(root)
        denom = dx2 + dz2
        
        a1 = a2 = -R*d.z - d.x*o.x - d.z*o.z
        a1 += root
        a2 -= root
        a1 /= denom
        a2 /= denom
        
        pt1 = addvv_(p1, multvs_(d, a1))
        pt2 = addvv_(p1, multvs_(d, a2))
        
        cz = self.z_height - self.radius
        
        if R >= 0:
            if pt1.z < cz:
                a1 = INF
            if pt2.z < cz:
                a2 = INF
        else:
            if pt1.z > cz:
                a1 = INF
            if pt2.z > cz:
                a2 = INF
        
        if is_base_ray:
            if not (<Shape>self.shape).point_inside_c( pt1.x, pt1.y ):
                a1 = INF
            if not (<Shape>self.shape).point_inside_c( pt2.x, pt2.y ):
                a2 = INF
        
        if a2 < a1:
            a1 = a2
        
        if a1>1.0 or a1<self.tolerance:
            return -1
        return a1 * sep_(p1, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        
        p.z -= (self.z_height - self.radius)
        if self.radius < 0:
            p.z = -p.z
            p.x = -p.x
        p.y = 0
        return norm_(p)
    
    cdef double eval_z_c(self, double x, double y) nogil:
        cdef:
            double R = self.radius
        if R>=0:
            return self.z_height + sqrt(R*R - x*x) - R
        else:
            return self.z_height - sqrt(R*R - x*x) - R
    
    
cdef class AxiconFace(ShapedFace):
    """
    While technically, we can use the conic surface to generate a cone, it requires setting some parameters to infinity which 
    is often inaccurate to compute.
    
    The gradient is the slope of the sides, dz/dr
    """
    cdef:
        public double z_height, gradient
        
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
        self.gradient = kwds.get('gradient', 0.0)
        
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            double a1, a2, root, ox2, oy2, oz2, dx2, dy2, dz2, beta2, denom
            double beta = self.gradient
            vector_t o, d, pt1, pt2
            
        d = subvv_(p2, p1)
        o = p1
        o.z -= self.z_height
        
        beta2 = beta*beta
        ox2 = o.x*o.x
        oy2 = o.y*o.y
        oz2 = o.z*o.z
        dx2 = d.x*d.x
        dy2 = d.y*d.y
        dz2 = d.z*d.z
        
        root = -beta2*dx2*oy2 + 2*beta2*d.x*d.y*o.x*o.y - beta2*dy2*ox2 + dx2*oz2 - 2*d.x*d.z*o.x*o.z + \
                dy2*oz2 - 2*d.y*d.z*o.y*o.z + dz2*ox2 + dz2*oy2
        denom = (beta2*dx2 + beta2*dy2 - dz2)
                
        if root < 0: #no intersection
            return -1
        
        root = beta*sqrt(root)
        
        a1 = -beta2*d.x*o.x - beta2*d.y*o.y + d.z*o.z
        a2 = a1 + root
        a1 -= root
        a1 /= denom
        a2 /= denom
        
        pt1 = addvv_(p1, multvs_(d, a1))
        pt2 = addvv_(p1, multvs_(d, a2))
        
        if pt1.z > self.z_height:
            a1 = INF
        if pt2.z > self.z_height:
            a2 = INF
        
        if is_base_ray:
            if not (<Shape>self.shape).point_inside_c( pt1.x, pt1.y ):
                a1 = INF
            if not (<Shape>self.shape).point_inside_c( pt2.x, pt2.y ):
                a2 = INF
        
        if a2 < a1:
            a1 = a2
        
        if a1>1.0 or a1<self.tolerance:
            return -1
        return a1 * sep_(p1, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef:
            double r, beta
            
        beta = self.gradient
        r = sqrt(p.x*p.x + p.y*p.y)
        p.z = 1.0
        p.x = beta * p.x/r
        p.y = beta * p.y/r
        return p
    
    cdef double eval_z_c(self, double x, double y) nogil:
        return self.z_height - (self.gradient * sqrt(x*x + y*y))
    
    
cdef double intersect_conic(vector_t a, vector_t d, double curvature, double conic_const):
    cdef:
        double beta = 1 + conic_const
        double R = -curvature
        double A,B,C,D
        #double pt1, pt2
    
    #print("inputs:", a.x, a.y, a.z, d.x, d.y, d.z, beta, R)
    ###Obtained by sympy evaluation of the equations
    ### Arranged into quadratic form i.e. A*(alpha**2) + B*alpha + C = 0
    A = beta**2*d.z**2 + beta*d.x**2 + beta*d.y**2
    B = -2*R*beta*d.z + 2*a.x*beta*d.x + 2*a.y*beta*d.y + 2*a.z*beta**2*d.z
    C = -2*R*a.z*beta + a.x**2*beta + a.y**2*beta + a.z**2*beta**2
    
    ##Get roots
    D = B*B - 4*A*C
    if D < 0: #no solution
        return -1
    
    D = sqrt(D)
    
### This turns out to not be necessary
#     #1st root
#     a1 = (-B+D)/(2*A) 
#     pt1 = a.z + d.z*a1 #addvv_(a, multvs_(d, a1))
#     #2nd root
#     a2 = (-B-D)/(2*A)
#     pt2 = a.z + d.z*a2 #addvv_(a, multvs_(d, a2))
#     
#     print("R.beta.dz:", R*beta*d.z, a1, pt1, a2, pt2, A, B, C, D)
#     if R*beta*d.z <= 0:
#         if pt1 < R/beta:
#             a1 = INF
#         if pt2 < R/beta:
#             a2 = INF
#     else:
#         if pt1 > R/beta:
#             a1 = INF
#         if pt2 > R/beta:
#             a2 = INF
#             
#     if a2 < a1:
#         return a2
#     else:
#         return a1
    
    #print("R.beta.dz:", R*beta*d.z)
    if R*beta*d.z <= 0:
        #1st root
        return (-B+D)/(2*A) 
    else:
        #2nd root
        return (-B-D)/(2*A)
    
         
    
cdef class ConicRevolutionFace(ShapedFace):
    """This is surface of revolution formed from a conic section. Spherical and ellipsoidal faces
    are a special case of this.
    
    curvature = radius of curvature
    """
    cdef:
        public double curvature, z_height, conic_const
    
    params = [] 
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
        self.conic_const = kwds.get('conic_const', 0.0)
        self.curvature = kwds.get('curvature', 10.0)
    
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          
        returns:
          the distance along the ray to the first valid intersection. No
          intersection can be indicated by a negative value.
        """
        cdef:
            double a1
            vector_t d, a, pt1
            
        d = subvv_(p2, p1) #the input ray direction, in local coords.
        a = p1
        a.z -= self.z_height
        
        a1 = intersect_conic(a, d, self.curvature, self.conic_const)
        
        pt1 = addvv_(a, multvs_(d, a1))
            
        if is_base_ray and not (<Shape>(self.shape)).point_inside_c(pt1.x, pt1.y):
            return INF
            
        if a1>1.0 or a1<self.tolerance:
            return -1
        
        return a1 * sep_(p1, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef:
            double R = -self.curvature
            double beta = 1 + self.conic_const
            vector_t g #output gradient vector
            int sign = -1 if self.invert_normals else 1
            
        p.z -= self.z_height
            
        g.z = 2*beta*(R-beta*p.z)
        g.x = - p.x * 2 * beta 
        g.y = - p.y * 2 * beta
        
        if (R*beta) < 0:
            sign *= -1
            
        g.z *= sign
        g.y *= sign
        g.x *= sign
        
        return norm_(g)
    
    cdef double eval_z_c(self, double x, double y) nogil:
        cdef:
            double r2 = (x*x) + (y*y)
            double R = self.curvature
            double R2 = R*R
        
        return self.z_height  - (r2/(R + sqrt(R2 - (1+self.conic_const)*r2)))
    
    cdef double eval_implicit_c(self, double x, double y, double z) nogil:
        return z - self.eval_z_c(x,y)
    
    
cdef struct aspheric_t:
    double R #curvature
    double beta #=1+conic const
    double A4
    double A6
    double A8
    double A10
    double A12
    double A14
    double A16
    vector_t a
    vector_t d
    
    
cdef double eval_aspheric_impf(aspheric_t A, double alpha):
    cdef: 
        double out
        double r2 = ((A.a.x + alpha*A.d.x)**2 + (A.a.y + alpha*A.d.y)**2)
    
    out = r2
    out /= A.R*(1 + sqrt(1 - A.beta*r2/(A.R**2)) )
    out -= A.a.z + alpha*A.d.z
    out += A.A4*(r2**2) + A.A6*(r2**3) + A.A8*(r2**4) + A.A10*(r2**5) + A.A12*(r2**6) + A.A14*(r2**7) + A.A16*(r2**8)
    return out


cdef double eval_aspheric_grad(aspheric_t A, double alpha):
    cdef:
        double out
        double r2 = ((A.a.x + alpha*A.d.x)**2 + (A.a.y + alpha*A.d.y)**2)
        double dx = A.d.x*(A.a.x + alpha*A.d.x)
        double dy = A.d.y*(A.a.y + alpha*A.d.y)
        
    out = A.A10*(10*dx + 10*dy) *(r2**4)
    out += A.A12*(12*dx + 12*dy) *(r2**5)
    out += A.A14*(14*dx + 14*dy) *(r2**6)
    out += A.A16*(16*dx + 16*dy) *(r2**7)
    out +=  A.A4*(4*dx + 4*dy)*(r2)  
    out +=  A.A6*(6*dx + 6*dy)*(r2**2)  
    out +=  A.A8*(8*dx + 8*dy)*(r2**3) - A.d.z  
    out += (2*dx + 2*dy)/(A.R*(sqrt(1 - A.beta*(r2)/A.R**2) + 1))  
    out += A.beta*(2*dx + 2*dy)*(r2)/(2*(A.R**3)*sqrt(1 - A.beta*r2/(A.R**2))*(sqrt(1 - A.beta*r2/(A.R**2)) + 1)**2)
    return out

    
cdef class AsphericFace(ShapedFace):
    """This is the general aspheric lens surface formula.
    
    curvature = radius of curvature
    """
    cdef:
        public double curvature, z_height, conic_const, A4, A6, A8, A10, A12, A14, A16
        public double atol
    
    params = [] 
    
    def __cinit__(self, **kwds):
        self.z_height = kwds.get('z_height', 0.0)
        self.conic_const = kwds.get('conic_const', 0.0)
        self.curvature = kwds.get('curvature', 25.0)
        self.A4 = kwds.get('A4',0.0)
        self.A6 = kwds.get('A6',0.0)
        self.A8 = kwds.get('A8',0.0)
        self.A10 = kwds.get('A10',0.0)
        self.A12 = kwds.get('A12',0.0)
        self.A14 = kwds.get('A14',0.0)
        self.A16 = kwds.get('A16',0.0)
        self.atol = kwds.get("atol", 1.0e-8)
        
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        """Intersects the given ray with this face.
        
        params:
          p1 - the origin of the input ray, in local coords
          p2 - the end-point of the input ray, in local coords
          
        returns:
          the distance along the ray to the first valid intersection. No
          intersection can be indicated by a negative value.
        """
        cdef:
            double a1, dz, f, f_last
            double tol = self.atol**2
            vector_t d, a, pt1
            aspheric_t A
            
        d = subvv_(p2, p1) #the input ray direction, in local coords.
        a = p1
        a.z -= self.z_height
        
        a1 = intersect_conic(a, d, self.curvature, self.conic_const)
        
        A.R = -self.curvature
        A.beta = 1 + self.conic_const
        A.A4 = self.A4
        A.A6 = self.A6
        A.A8 = self.A8
        A.A10 = self.A10
        A.A12 = self.A12
        A.A14 = self.A14
        A.A16 = self.A16
        A.a = a
        A.d = d

        ### Find root using Newton's method. Typically only a 3-4 iterations are required.
         
        f = f_last = eval_aspheric_impf(A, a1)
        dz = - f / eval_aspheric_grad(A, a1)
        #print("Start:", dz, a1)

        #### If we've not converged after 100 iterations, then it ain't going converge ever.
        for i in range(100):
            a1 += dz
            if dz*dz < tol:
                break
            f = eval_aspheric_impf(A, a1)
            if fabs(f) > fabs(f_last): #We're not converging
                return -1
            f_last = f
            dz = - f / eval_aspheric_grad(A, a1)
        else:
            return -1     
        
        #print("Converged:", dz, a1, i)
        
        pt1 = addvv_(a, multvs_(d, a1))

        if is_base_ray and not (<Shape>(self.shape)).point_inside_c(pt1.x, pt1.y):
            return INF
            
        if a1>1.0 or a1<self.tolerance:
            return -1
        
        #print("Ret:", a1 * sep_(p1, p2))
        return a1 * sep_(p1, p2)
    
    cdef vector_t compute_normal_c(self, vector_t p):
        """Compute the surface normal in local coordinates,
        given a point on the surface (also in local coords).
        """
        cdef:
            double R = -self.curvature
            double beta = 1 + self.conic_const
            vector_t g #output gradient vector
            int sign = -1 if self.invert_normals else 1
            double r2, r, df, root
            
        p.z -= self.z_height
        
        r2 = p.x*p.x + p.y*p.y
        r = sqrt(r2)
        root = sqrt(1-(beta*(r2)/(R*R)))
        df = 10*self.A10*(r2**4) + 8*self.A8*(r2**3) + 6*self.A6*(r2**2) + 4*self.A4*r2
        df += 16*self.A16*(r2**7) + 14*self.A14*(r2**6) + 12*self.A12*(r2**5) 
        df += 2/(R*(1+root))
        df += beta*(r2)/((R**3)*root*((1+root)**2))
            
        g.z = 1.0
        g.x = - df*p.x
        g.y = - df*p.y
        
        g.z *= sign
        g.y *= sign
        g.x *= sign
        
        return norm_(g)

    cdef double eval_z_c(self, double x, double y) nogil:
        cdef: 
            double out
            double r2 = (x*x) + (y*y)
            double R = self.curvature
    
        out = r2
        out /= (R + sqrt(R*R - (1+self.conic_const)*r2) )
        out += self.A4*(r2**2) + self.A6*(r2**3) + self.A8*(r2**4) + self.A10*(r2**5) + self.A12*(r2**6) + self.A14*(r2**7) + self.A16*(r2**8)
        out += self.z_height
        return out
    
    cdef double eval_implicit_c(self, double x, double y, double z) nogil:
        return z - self.eval_z_c(x,y)
    
    
cdef class DistortionFace(ShapedFace):
    """This class wraps another ShapedFace object, and applies a small distortion to it's
    surface geometry. The distortion is given by an instance of a Distortion subclass
    """
    cdef:
        public ShapedFace base_face
        public Distortion distortion
        public double accuracy
        
    def __cinit__(self, **kwds):
        face = kwds.get("base_face")
        self.base_face = face
        self.distortion = kwds.get("distortion")
        self.shape = kwds.get('shape', face.shape)
        self.accuracy = kwds.get("accuracy", 1e-6)
        
    cdef double intersect_c(self, vector_t p1, vector_t p2, int is_base_ray):
        cdef:
            ShapedFace face=self.base_face
            double z_shift=0.0, tolerance, a1, a2, h=sep_(p2,p1)
            vector_t dxdyz, pt1, d, n, o, q1, q2
            int i
            Distortion dist=self.distortion
            Shape shape = self.shape
            
        tolerance = self.accuracy
        
        a2 = face.intersect_c(p1,p2,0) #Don't check the shape yet
        if a2>h or a2<self.tolerance: #If no intersection, then, no intersection
            #print("NO intersection", a2, p1.z, p2.z)
            return -1.0
        
        #print("start estimate:", a2)
        #starting estimate of intersection
        d = subvv_(p2,p1)
        pt1 = addvv_(p1, multvs_(d,a2/h))
        dxdyz = dist.z_offset_and_gradient_c(pt1.x, pt1.y)
        
        n = face.compute_normal_c(pt1)
        #print("base normal:", n.x, n.y, n.z)
        #print("pt1:", pt1.x, pt1.y, pt1.z)
        #print("dx:", dxdyz.x, "dy:", dxdyz.y, "z:", dxdyz.z)
        
        pt1.z += dxdyz.z
        
        n.x /= n.z
        n.y /= n.z
        n.x -= dxdyz.x
        n.y -= dxdyz.y
        
        o=subvv_(p1, pt1)
        ### Better estimate for a
        a1 = -h*dotprod_(o,n)/dotprod_(d,n)
        #print("better estimate", a1)
        if a1 < fabs(dxdyz.z):
            #print("self-intersection. bailing.")
            return -1.0
        
        for i in range(20):
            pt1 = addvv_(p1, multvs_(d,a1/h))
            #print("a_error:", a1-a2, i)
            if fabs(a1 - a2) < tolerance:
                break
            z_shift = dist.z_offset_c(pt1.x, pt1.y)
            
            q1 = p1
            q2 = p2
            q1.z -= z_shift
            q2.z -= z_shift
            a2 = face.intersect_c(q1, q2, 0)
            
            pt1 = addvv_(q1, multvs_(d,a2/h)) #this point on base
            n = face.compute_normal_c(pt1)
            dxdyz = dist.z_offset_and_gradient_c(pt1.x, pt1.y)
            pt1.z += dxdyz.z #move to distorted face
            
            n.x /= n.z
            n.y /= n.z
            n.x -= dxdyz.x
            n.y -= dxdyz.y
            
            o=subvv_(p1, pt1)
            ### Better estimate for a
            a2 = a1
            a1 = -h*dotprod_(o,n)/dotprod_(d,n)
            
        if not shape.point_inside_c(pt1.x, pt1.y):
            return -1.0
        #print("Final a1:", a1)
        return a1
            
    cdef vector_t compute_normal_c(self, vector_t p):
        cdef:
            vector_t n, p1, dxdyz

        dxdyz = self.distortion.z_offset_and_gradient_c(p.x, p.y)
        p1 = p
        p1.z -= dxdyz.z
        n = self.base_face.compute_normal_c(p1)
        n.x /= n.z
        n.y /= n.z
        n.z = 1.0
        n.x -= dxdyz.x
        n.y -= dxdyz.y
        return norm_(n)
            
    cdef double eval_z_c(self, double x, double y) nogil:
        cdef:
            double z
        z = self.base_face.eval_z_c(x,y)
        z += self.distortion.z_offset_c(x,y)
        return z
    


#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from traits.api import Float, Instance, on_trait_change, Array, Property,\
        cached_property, List, observe, Tuple, cached_property, Str, Complex

from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, Group, ListEditor

from tvtk.api import tvtk
     
from raypier.bases import Optic, normaliseVector, NumEditor,\
    ComplexEditor, Traceable, transformPoints, transformNormals,\
    ShapedOptic
    
from raypier.core.cfaces import CircularFace, SphericalFace, ConicRevolutionFace, AsphericFace
from raypier.core.ctracer import FaceList
from raypier.core.cmaterials import DielectricMaterial, CoatedDispersiveMaterial, PECMaterial, SingleLayerCoatedMaterial
from raypier.vtk_algorithms import EmptyGridSource
from raypier.shapes import CircleShape, BaseShape

import math, numpy
import itertools

def pairwise(itr):
    a1,a2 = itertools.tee(itr)
    next(a2)
    return zip(a1,a2)


class BaseLens(Optic):
    abstract = True


class PlanoConvexLens(BaseLens):
    abstract = False
    name = "Plano-Convex Lens"
    
    CT = Float(5.0, desc="centre thickness")
    diameter = Float(15.0)
    offset = Float(0.0)
    curvature = Float(11.7, desc="radius of curvature for spherical face")    
    
    n_coating = Complex(1.3+0.0j)
    coating_thickness = Float(0.25)
    
    vtk_grid = Instance(EmptyGridSource, ())
    vtk_cylinder = Instance(tvtk.Cylinder, ())
    vtk_sphere = Instance(tvtk.Sphere, ())
    
    clip2 = Instance(tvtk.ClipDataSet, ())
        
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('n_inside'),
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature', editor=NumEditor),
                       Item('n_coating'),
                       Item('coating_thickness', editor=NumEditor)
                       )
                    )
    
    def _material_default(self):
        #return BaseLens._material_default(self)
        return SingleLayerCoatedMaterial(n_inside=self.n_inside,
                                         n_outside=self.n_outside,
                                         n_coating=self.n_coating, 
                                         coating_thickness=self.coating_thickness)
        
    def _n_coating_changed(self, vnew):
        self.material.n_coating = vnew
        self.update = True
        
    def _coating_thickness_changed(self, vnew):
        self.material.coating_thickness = vnew
        self.update=True
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, diameter=self.diameter,
                                material = self.material), 
                SphericalFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                z_height=self.CT, curvature=self.curvature)]
        return fl
    
    def _CT_changed(self, new_ct):
        self.faces.faces[1].z_height = new_ct
        
    def _curvature_changed(self, new_curve):
        self.faces.faces[1].curvature = new_curve
    
    def make_step_shape(self):
        from .step_export import make_spherical_lens
        shape = make_spherical_lens(self.CT, self.diameter, self.curvature, 
                                   self.centre, self.direction, self.x_axis)
        return shape, "blue1"
    
    @on_trait_change("CT, diameter, curvature")
    def config_pipeline(self):
        ct = self.CT
        rad = self.diameter/2
        curve = self.curvature
        
        size = 41
        spacing = 2*rad / (size-1)
        if curve >= 0.0:
            extra=0.0
        else:
            extra = - curve - math.sqrt(curve**2 - rad**2)
        lsize = int((ct+extra)/spacing) + 2
        
        grid = self.vtk_grid
        grid.dimensions = (size,size,lsize)
        grid.origin = (-rad, -rad, 0)
        grid.spacing = (spacing, spacing, spacing)
                      
        cyl = self.vtk_cylinder
        cyl.center = (0,0,0)
        cyl.radius = rad
        
        s = self.vtk_sphere
        s.center = (0,0,ct - curve)
        s.radius = abs(curve)
        
        self.clip2.inside_out = bool(curve >= 0.0)
        
        self.vtk_grid.modified()
        self.update=True
        
                                         
    def _pipeline_default(self):
        grid = self.vtk_grid
        #grid.set_execute_method(self.create_grid)
        grid.modified()
        
        trans = tvtk.Transform()
        trans.rotate_x(90.)
        cyl = self.vtk_cylinder
        cyl.transform = trans
        
        clip1 = tvtk.ClipVolume(input_connection=grid.output_port,
                                 clip_function=self.vtk_cylinder,
                                 inside_out=1)
        
        self.clip2.set(input_connection = clip1.output_port,
                      clip_function=self.vtk_sphere,
                      inside_out=1)
        
        topoly = tvtk.GeometryFilter(input_connection=self.clip2.output_port)
        norms = tvtk.PolyDataNormals(input_connection=topoly.output_port)
        
        transF = tvtk.TransformFilter(input_connection=norms.output_port, 
                                      transform=self.transform)
        self.config_pipeline()
        grid.modified()
        return transF
    
    
class ShapedPlanoSphericLens(ShapedOptic):
    abstract = False
    name = "Plano-Convex Lens"
    
    CT = Float(5.0, desc="centre thickness")
    diameter = Float(15.0)
    offset = Float(0.0)
    curvature = Float(11.7, desc="radius of curvature for spherical face")
    
    shape = Instance(BaseShape, ())
    
    grid_extent = Property(depends_on="CT, diameter")
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('n_inside'),
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature', editor=NumEditor)
                       )
                    )
    
    def _shape_default(self):
        s1 = CircleShape(radius=self.diameter/2.)
        return s1
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, diameter=self.diameter,
                                material = self.material), 
                SphericalFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                z_height=self.CT, curvature=self.curvature)]
        return fl
    
    @cached_property
    def _get_grid_extent(self):
        r = self.diameter/2
        return (-r,r,-r,r,0,self.CT)
    
    def _CT_changed(self, new_ct):
        self.faces.faces[1].z_height = new_ct
        self.update = True
        
    def _curvature_changed(self, new_curve):
        self.faces.faces[1].curvature = new_curve
        self.grid_in.modified()
        self.update = True
        
    def _diameter_changed(self, dnew):
        self.shape.radius = dnew/2.
        self.grid_in.modified()
        
    def eval_sag_top(self, points):
        centre = numpy.array([0,0,-(self.curvature-self.CT)])
        func = ((points - centre)**2).sum(axis=1) - (self.curvature**2)
        return func
    
    def eval_sag_bottom(self, points):
        return None
    
    
class SurfaceOfRotationLens(BaseLens):
    ###An array representing the 2D profile of the lens to be revolved
    profile = Array(shape=(None,2), dtype=numpy.double, transient=True)
    
    data_source = Instance(tvtk.ProgrammableSource, (), transient=True)
    
    revolve = Instance(tvtk.RotationalExtrusionFilter, (), 
                       {'capping': False,
                        'angle': 360.0,
                        'resolution': 60 
                        },
                        transient=True)
    
    def _pipeline_default(self):
        self.config_profile()
        
        source = self.data_source
        def execute():
            yz = self.profile
            x = numpy.zeros(yz.shape[0])
            points = numpy.column_stack((x,yz))
            
            cells = [list(range(len(x))),]
            
            output = source.poly_data_output
            output.points = points
            output.polys = cells
        source.set_execute_method(execute)
        print("Made PIPELINE")
        
        revolve = self.revolve
        revolve.input_connection = source.output_port
        
        t = self.transform
        transf = tvtk.TransformFilter(input_connection=revolve.output_port, 
                                      transform=t)
        return transf
    
    def _profile_changed(self):
        self.data_source.modified()
        self.faces.faces = self.make_faces()
        self.update=True
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = self.make_faces()
        return fl
    
    def config_profile(self):
        pass
    
    def make_faces(self):
        return []
    
    
class PlanoConicLens(SurfaceOfRotationLens):
    abstract = False
    name = "Plano-Conic Section Lens"
    
    CT = Float(5.0, desc="centre thickness")
    diameter = Float(15.0)
    offset = Float(0.0) #Fixed at zero
    conic_const = Float(4.0, desc="conic constant")
    curvature = Float(11.7, desc="radius of curvature for spherical face")
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('n_inside'),
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature', editor=NumEditor),
                       Item('conic_const', editor=NumEditor)
                       )
                    )
    
    def make_faces(self):
        fl = [CircularFace(owner=self, diameter=self.diameter,
                                material = self.material), 
                ConicRevolutionFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                conic_const=self.conic_const,
                                z_height=self.CT, curvature=self.curvature)]
        return fl
    
    def _CT_changed(self, new_ct):
        self.faces.faces[1].z_height = new_ct
        
    def _curvature_changed(self, new_curve):
        self.faces.faces[1].curvature = new_curve
        
    def _conic_const_changed(self, new_conic):
        self.faces.faces[1].conic_const = new_conic
        
    @on_trait_change("CT, diameter, conic_const, curvature")
    def config_profile(self):
        n_segments = 20 #number of lines in the curved section
        r_max = self.diameter/2 #maximum radius
        k = self.conic_const
        R = self.curvature
        r = numpy.linspace(0,r_max, n_segments)
        
        z = self.CT - (r**2)/(R*(1 + numpy.sqrt(1 - (1+k)*(r**2)/(R**2))))
        
        profile = list(zip(r,z))
        profile.extend([(r_max, 0.0), (0.0,0.0)])
        self.profile = profile
        
    
    
    
class BiConicLens(PlanoConicLens):
    name = "Bi-Conic Section Lens"
    
    conic_const2 = Float(-4.0, desc="conic constant, lower face")
    curvature2 = Float(11.7, desc="radius of curvature for spherical face")
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('n_inside'),
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       Item('curvature', editor=NumEditor),
                       Item('conic_const', editor=NumEditor),
                       Item('curvature2', editor=NumEditor),
                       Item('conic_const2', editor=NumEditor)
                       )
                    )
    
    def _curvature2_changed(self, new_curve):
        self.faces.faces[0].curvature = new_curve
        
    def _conic_const2_changed(self, new_conic):
        self.faces.faces[0].conic_const = new_conic
    
    def make_faces(self):
        fl = [ConicRevolutionFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                conic_const=self.conic_const2,
                                z_height=0.0, curvature=self.curvature2,
                                invert_normals=True), 
                ConicRevolutionFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                conic_const=self.conic_const,
                                z_height=self.CT, curvature=self.curvature)]
        return fl

    @on_trait_change("CT, diameter, conic_const, curvature, conic_const2, curvature2")
    def config_profile(self):
        n_segments = 20 #number of lines in the curved section
        r_max = self.diameter/2 #maximum radius
        k = self.conic_const
        k2 = self.conic_const2
        R = self.curvature
        R2 = self.curvature2
        r = numpy.linspace(0,r_max, n_segments)
        
        #top surface
        z = self.CT - (r**2)/(R*(1 + numpy.sqrt(1 - (1+k)*(r**2)/(R**2))))
        z2 = - (r**2)/(R2*(1 + numpy.sqrt(1 - (1+k2)*(r**2)/(R2**2))))
        
        profile = list(zip(r,z))
        profile2 = list(zip(r,z2))
        
        profile.extend(profile2[-1::-1])
        self.profile = profile
        
        
class AsphericLens(SurfaceOfRotationLens):
    abstract = False
    name = "Bi-Aspheric Lens"
    
    CT = Float(5.0, desc="centre thickness")
    diameter = Float(25.0)
    offset = Float(0.0) #Fixed at zero
    
    A_curvature = Float(-25.0)
    A_conic = Float(-0.0)
    A4 = Float(0.0)
    A6 = Float(0.0)
    A8 = Float(0.0)
    A10 = Float(0.0)
    A12 = Float(0.0)
    A14 = Float(0.0)
    A16 = Float(0.0)
    
    B_curvature = Float(25.0)
    B_conic = Float(-0.0)
    B4 = Float(0.0)
    B6 = Float(0.0)
    B8 = Float(0.0)
    B10 = Float(0.0)
    B12 = Float(0.0)
    B14 = Float(0.0)
    B16 = Float(0.0)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,  
                       Item('n_inside'),
                       Item('CT', editor=NumEditor),
                       Item('diameter', editor=NumEditor),
                       VGroup(
                           Item('A_curvature', editor=NumEditor),
                           Item('A_conic', editor=NumEditor),
                           Item('A4', editor=NumEditor),
                           Item('A6', editor=NumEditor),
                           Item('A8', editor=NumEditor),
                           Item('A10', editor=NumEditor),
                           label="A Surface",
                           show_border = True
                           ),
                       VGroup(
                           Item('B_curvature', editor=NumEditor),
                           Item('B_conic', editor=NumEditor),
                           Item('B4', editor=NumEditor),
                           Item('B6', editor=NumEditor),
                           Item('B8', editor=NumEditor),
                           Item('B10', editor=NumEditor),
                           label="B Surface",
                           show_border = True
                           ),
                       )
                    )
    
#     def make_faces(self):
#         fl = [ConicRevolutionFace(owner=self, diameter=self.diameter,
#                                 material=self.material,
#                                 conic_const=self.A_conic,
#                                 z_height=0.0, curvature=self.A_curvature,
#                                 invert_normals=True), 
#                 ConicRevolutionFace(owner=self, diameter=self.diameter,
#                                 material=self.material,
#                                 conic_const=self.B_conic,
#                                 z_height=self.CT, curvature=self.B_curvature)]
#         return fl
    
    def make_faces(self):
        fl = [AsphericFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                conic_const=self.A_conic,
                                z_height=0.0, curvature=self.A_curvature,
                                A4=-self.A4, A6=-self.A6, A8=-self.A8, A10=-self.A10,
                                A12=-self.A12, A14=-self.A14, A16=-self.A16,
                                invert_normals=True,
                                atol=1e-16), 
                AsphericFace(owner=self, diameter=self.diameter,
                                material=self.material,
                                conic_const=self.B_conic,
                                z_height=self.CT, curvature=self.B_curvature,
                                A4=-self.B4, A6=-self.B6, A8=-self.B8, A10=-self.B10,
                                A12=-self.B12, A14=-self.B14, A16=-self.B16,
                                atol=1e-16)]
        return fl
    
    def eval_A(self, y):
        y2=numpy.asarray(y)**2
        R = self.A_curvature
        beta = (1 + self.A_conic)
        z = self.A4*(y2**2) + self.A6*(y2**3) + self.A8*(y2**4) + self.A10*(y2**5) + y2 / (R*(1 + numpy.sqrt(1 - beta*y2/(R**2))))
        return -z
    
    def eval_B(self, y):
        y2=numpy.asarray(y)**2
        R = self.B_curvature
        beta = (1 + self.B_conic)
        z = self.B4*(y2**2) + self.B6*(y2**3) + self.B8*(y2**4) + self.B10*(y2**5) + y2 / (R*(1 + numpy.sqrt(1 - beta*y2/(R**2))))
        return self.CT - z
    
    @on_trait_change("CT, diameter, A_conic, A_curvature, A4, A6, A8, A10, B_conic, B_curvature, B4, B6, B8, B10")
    def config_profile(self):
        n_segments = 20 #number of lines in the curved section
        r_max = self.diameter/2 #maximum radius
        r = numpy.linspace(0,r_max, n_segments)
        
        profile = list(zip(r, self.eval_A(r)))
        profile2 = list(zip(r, self.eval_B(r)))
        
        profile.extend(profile2[-1::-1])
        self.profile = profile
    
    

    
    


        
if __name__=="__main__":
    from raypier.tracer import RayTraceModel
    from raypier.sources import ConfocalRaySource
    
    lens = PlanoConvexLens(orientation=0.0,
                           elevation=0.0,
                           CT=5.,
                           curvature=12.)
    
    source = ConfocalRaySource(focus=(0,0,-30),
                            direction=(0,0,1),
                            working_dist = 0.1,
                            number=20,
                            detail_resolution=5,
                            theta=10.,
                            scale_factor=0.1)
    
    model = RayTraceModel(optics=[lens], 
                          sources=[source,])
    model.configure_traits()




from traitsui.api import TextEditor, TupleEditor


NumEditor = TextEditor(auto_set=False, enter_set=True, evaluate=float)
IntEditor = TextEditor(auto_set=False, enter_set=True, evaluate=int)
ComplexEditor = TextEditor(auto_set=False, enter_set=True, 
                           evaluate=float)

ROField = TextEditor(auto_set=False, enter_set=True, evaluate=float)
VectorEditor = TupleEditor(labels=['x','y','z'], auto_set=False, enter_set=True)



import faulthandler
faulthandler.enable()

import unittest
import numpy

from matplotlib import pyplot as pp
from collections import Counter


from raypier.core.cdistortions import SimpleTestZernikeJ7, eval_nmk, zernike_R, zernike_Rprime, \
        ZernikeDistortion, zernike_R_over_r
from raypier.core import cdistortions


class TestSimpleTestZernikeJ7(unittest.TestCase):
    def test_gradient_and_sag(self):
        dist = SimpleTestZernikeJ7(unit_radius=10.0, amplitude=1.0)
        
        x_ = numpy.linspace(-10.,10.,200)
        y_ = numpy.linspace(-10.,10.,200)
        
        x,y = numpy.meshgrid(x_,y_)
        
        v = dist.z_offset_and_gradient(x.reshape(-1),y.reshape(-1))
        v = numpy.asarray(v)
        pp.imshow(v[:,2].reshape(x.shape))
        
        pp.show()
        
        
### Copied off Wikipedia
zernike_R_funcs = {
    (0,0): lambda r: 1.0,
    (1,1): lambda r: r,
    (2,0): lambda r: 2*(r**2) - 1,
    (2,2): lambda r: r**2,
    (3,1): lambda r: 3*(r**3) - (2*r),
    (3,3): lambda r: r**3,
    (4,0): lambda r: 6*(r**4) - 6*(r**2) + 1,
    (4,2): lambda r: 4*(r**4) - 3*(r**2),
    (4,4): lambda r: r**4,
    (5,1): lambda r: 10*(r**5) - 12*(r**3) + 3*r,
    (5,3): lambda r: 5*(r**5) - 4*(r**3),
    (5,5): lambda r: r**5,
    (6,0): lambda r: 20*(r**6) - 30*(r**4) + 12*(r**2) - 1,
    (6,2): lambda r: 15*(r**6) - 20*(r**4) + 6*(r**2),
    (6,4): lambda r: 6*(r**6) - 5*(r**4),
    (6,6): lambda r: r**6
    }


class TestEvalNMK(unittest.TestCase):
    def test_eval_nmk(self):
        n,m,k = eval_nmk(19)
        nmk = cdistortions.jnm_map
        
        ### Taken from  https://en.wikipedia.org/wiki/Zernike_polynomials
        ### ANSI standard J indexing
        ### table of (J,N,M)
        check = [(0,0,0),
                 (1,1,-1),
                 (2,1,1),
                 (3,2,-2),
                 (4,2,0),
                 (5,2,2),
                 (6,3,-3),
                 (7,3,-1),
                 (8,3,1),
                 (9,3,3),
                 (10,4,-4),
                 (11,4,-2),
                 (12,4,0),
                 (13,4,2),
                 (14,4,4),
                 (15,5,-5),
                 (16,5,-3),
                 (17,5,-1),
                 (18,5,1),
                 (19,5,3)
                  ]
        
        tab = [(i,n,m) for i,(n,m,k) in enumerate(nmk)]
        self.assertEqual(tab, check)
        
        tab = [(i,n,m) for i,(n,m,k) in enumerate(eval_nmk(i) for i in range(20))]
        self.assertEqual(tab, check)
        
        ###check there are no duplicate k-values for each (n, abs(m)) key.
        dups = {(n, abs(m)): k for n,m,k in nmk}
        c=Counter(dups.values())
        self.assertLessEqual(max(c.values()), 1)
        
        

        
        
class TestZernikeSeriesDistortion(unittest.TestCase):
    def test_eval_Z(self):
        Z = ZernikeDistortion(j7=1.0, unit_radius=10.0)
        Z2 = SimpleTestZernikeJ7(unit_radius=10.0, amplitude=1.0)
        
        x_ = numpy.linspace(-5.,5.,4)
        x,y = numpy.meshgrid(x_,x_)
        x=x.ravel()
        y=y.ravel()
        a = Z.z_offset(x,y)
        b = Z2.z_offset(x,y)
        self.assertTrue(numpy.allclose(a, b))
        del Z
        
    def test_show_num_threads(self):
        print(cdistortions.num_threads)
        
    def test_get_coefs(self):
        Z = ZernikeDistortion(j7=1.0, unit_radius=10.0)
        self.assertEqual(1.0, Z.get_coef(7))
        
    def test_eval_gradient(self):
        Z = ZernikeDistortion(j7=1.0, unit_radius=10.0)
        Z2 = SimpleTestZernikeJ7(unit_radius=10.0, amplitude=1.0)
        
        x_ = numpy.linspace(-5.,5.,4)
        x,y = numpy.meshgrid(x_,x_)
        x=x.ravel()
        y=y.ravel()
        a = Z.z_offset_and_gradient(x,y)
        b = Z2.z_offset_and_gradient(x,y)
        print(a)
        print(b)
        self.assertTrue(numpy.allclose(a, b))
        del Z
        
    def test_gradient_at_origin(self):
        Z = ZernikeDistortion(j7=1.0, unit_radius=10.0)
        Z2 = SimpleTestZernikeJ7(unit_radius=10.0, amplitude=1.0)
        x = y = numpy.array([0.0],'d')
        a = Z.z_offset_and_gradient(x,y)
        b = Z2.z_offset_and_gradient(x,y)
        print(a,b)
        self.assertTrue(numpy.allclose(a, b))
        
        
        
class TestEvalZernikeR(unittest.TestCase):
    def test_eval_zernike(self):
        r = 0.7071067811865476
        n,m,k = eval_nmk(7)
        k_max = max(eval_nmk(j)[2] for j in range(7))
        print("k_max=", k_max)
        workspace = numpy.zeros((3,k_max),'d')
        workspace[0,:] = float("nan")
        workspace[0,0] = 1.0
        print("k=",k, "n=",n, "m=",m)
        print("init=",workspace)
        R = zernike_R(r,k,n,m,workspace)
        print("R=",R)
        print("final=",workspace)
        
        R2 = 3*(r**3) - (2*r)
        self.assertAlmostEqual(R, R2)
        
    def test_eval_zernike_R_all(self):
        r_all = numpy.linspace(0,1.0,20)
        k_max = 3*4 + 6 + 1
        workspace = numpy.zeros((3,k_max), 'd')
        for r in r_all: 
            workspace[:,:] = float("nan")
            for n in range(7):
                if 2*(n//2) == n:
                    start=0
                else:
                    start=1
                for m in range(start,n+1,2):
                    half_n = n//2
                    k=half_n*(half_n+1) + m
                    R = zernike_R(r,k,n,m,workspace)
                    R2 = zernike_R_funcs[(n,m)](r)
                    self.assertAlmostEqual(R,R2)
                
        
    def test_R_over_r(self):
        r=0.13
        n,m,k = eval_nmk(19)
        k_max = (6//2)*(6//2 + 1) + 6 + 1
        print(">> k_max=", k_max)
        k_max *= 2
        workspace = numpy.zeros((3,k_max),'d')
        workspace[:,:] = float("nan")
        #workspace[0,0] = 1.0

        for n in range(1,7):
            if 2*(n//2)==n:
                start=2
            else:
                start=1
            for m in range(start,n+1,2):
                half_n = int(n//2)
                k = half_n*(half_n+1) + abs(m)
                R = zernike_R(r,k,n,m,workspace)
                #Ra = zernike_R2(r,k,n,m,workspace)
                R2 = r*zernike_R_over_r(r,k,n,m,workspace)
                print("n,m,k=", (n,m,k), "R=",R, "R2=", R2)
                self.assertAlmostEqual(R,R2)
        
        


#    Copyright 2009, Teraview Ltd., Bryan Cole
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from traits.api import Float, Instance, on_trait_change, Property, cached_property

from traitsui.api import View, Item, VGroup

from tvtk.api import tvtk


from raypier.bases import Traceable, normaliseVector, NumEditor,\
     ComplexEditor, VectorEditor, Optic, ShapedTraceable
     
from raypier.utils import transformPoints, dotprod
from raypier.sources import RayCollection
from raypier.core.cfaces import CircularFace, RectangularFace, ShapedSphericalFace
from raypier.core.ctracer import FaceList
from raypier.core.cmaterials import DielectricMaterial, \
        CoatedDispersiveMaterial, PECMaterial
from raypier.dispersion import BaseDispersionCurve, NondispersiveCurve
from raypier.shapes import CircleShape, RectangleShape

import math, numpy



class BaseMirror(Traceable):
    abstract=True
    def _vtkproperty_default(self):
        return tvtk.Property(opacity=1.0,
                             color=(0.8,0.8,0.8),
                                representation="surface")


class PECMirror(BaseMirror):
    name = "PEC Mirror"
    abstract = False
    diameter = Float(25.4)
    thickness = Float(5.0, desc="purely for visualisation purposes")
    offset = Float(0.0)
    
    vtk_cylinder = Instance(tvtk.CylinderSource, (),
                            dict(resolution=32), transient=True)
    
    cyl_trans = Instance(tvtk.Transform, (), transient=True)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor)
                        ),
                   )
    
    def _material_default(self):
        return PECMaterial()
    
    def _faces_default(self):
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, material=m)]
        return fl
    
    def make_step_shape(self):
        from raypier.step_export import make_cylinder
        cyl = make_cylinder(self.centre, 
                             self.direction, 
                             self.diameter/2, 
                             self.thickness,
                             self.offset,
                             self.x_axis)
        return cyl, "green"
    
    @on_trait_change("diameter, thickness, offset")
    def config_pipeline(self):
        cyl = self.vtk_cylinder
        cyl.radius = self.diameter/2.
        thick = self.thickness
        cyl.height = thick
        
        t = self.cyl_trans
        t.identity()
        t.translate(self.offset,0,thick/2.)
        t.rotate_x(90.)
        
        self.update = True
    
    def _pipeline_default(self):
        cyl = self.vtk_cylinder
        norms = tvtk.PolyDataNormals(input_connection=cyl.output_port)
        transF1 = tvtk.TransformFilter(input_connection=norms.output_port, 
                                       transform=self.cyl_trans)
        transF2 = tvtk.TransformFilter(input_connection=transF1.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return transF2
    
    
class SphericalMirrorWithHole(ShapedTraceable):
    name = "Spherical Mirror + Hole"
    abstract = False
    
    diameter = Float(25.4)
    hole_diameter = Float(5.0)
    hole_offset = Float(0.0)
    
    curvature = Float(100.0)
    
    thickness = Float(5.0, desc="centre thickness, purely for visualisation purposes")
    
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                       Item('hole_diameter', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('curvature', editor=NumEditor),
                       Item('hole_offset', editor=NumEditor)
                        ),
                   )
        
    
    def _material_default(self):
        return PECMaterial()
    
    def _shape_default(self):
        return CircleShape(radius=self.diameter/2.) ^ CircleShape(radius=self.hole_diameter/2.,
                                                                  centre=(self.hole_offset,0))
    
    def _faces_default(self):
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [ShapedSphericalFace(owner=self, material=m, shape=self.shape.cshape)]
        return fl
    
    @on_trait_change("diameter, hole_diameter, thickness, curvature, hole_offset")
    def config_params(self):
        self.grid_extent = self.eval_grid_extent()
        self.shape.shape1.radius = self.diameter/2.
        self.shape.shape2.radius = self.hole_diameter/2.
        self.shape.shape2.centre = (self.hole_offset,0)
        face = self.faces.faces[0]
        face.z_height = self.thickness
        face.curvature = -self.curvature
        self.update = True
        
    def eval_grid_extent(self):
        r = self.diameter/2
        c = self.curvature
        h = c - numpy.sqrt(c**2 - r**2)
        return (-r,r,-r,r,0,self.thickness + h)
    
    def eval_sag_top(self, points):
        centre = numpy.array([0,0,(self.curvature+self.thickness)])
        func = ((points - centre)**2).sum(axis=1) - (self.curvature**2)
        return -func
    
    def eval_sag_bottom(self, points):
        return None
    
    

class PlanarWindow(PECMirror, Optic):
    n_inside = 1.5
    name = "Planar window"
    abstract = False
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor),
                       Item('n_inside', editor=NumEditor),
                        ),
                   )
                   
    vtkproperty = tvtk.Property(opacity = 0.4,
                             color = (0.8,0.8,1.0))
                             
    def _material_default(self):
        return Optic._material_default(self)
                   
    def _thickness_changed(self, new_t):
        self.faces[1].z_plane = new_t
    
    def _faces_default(self):
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, z_plane=0, material=m),
                    CircularFace(owner=self, z_plane=self.thickness, 
                        invert_normal=True, material=m)]
        return fl
    

class PlanarDispersiveWindow(PECMirror, Optic):
    name = "Planar dispersive window"
    abstract = False
    
    material = Instance(CoatedDispersiveMaterial, ())
    
    dispersion = Instance(BaseDispersionCurve, desc="Dispersion curve for the glass")
    dispersion_coating = Instance(BaseDispersionCurve, desc="Dispersion curve for the coating")
    coating_thickness = Float(0.25, desc="Thickness of the AR coating, in microns")
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('diameter', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor),
                       Item('n_inside', editor=NumEditor),
                        ),
                   )
                   
    vtkproperty = tvtk.Property(opacity = 0.4,
                             color = (0.6,0.9,0.5))
                             
    @on_trait_change("dispersion, dispersion_coating, coating_thickness")
    def on_materials_changed(self):
        self.apply_materials()
        self.update = True
    
    def apply_materials(self):
        air = NondispersiveCurve(1.0)
        self.material.dispersion_inside = self.dispersion
        self.material.dispersion_outside = air
        self.material.dispersion_coating = self.dispersion_coating
        self.material.coating_thickness = self.coating_thickness
                   
    def _thickness_changed(self, new_t):
        self.faces[1].z_plane = new_t
        self.update = True
    
    def _faces_default(self):
        self.apply_materials()
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, z_plane=0, material=m),
                    CircularFace(owner=self, z_plane=self.thickness, 
                        invert_normal=True, material=m)]
        return fl


class RectMirror(BaseMirror):
    name = "Rectangular Mirror"
    abstract = False
    length = Float(25.4)
    width = Float(25.4)
    thickness = Float(5.0, desc="purely for visualisation purposes")
    offset = Float(0.0)
    
    vtk_cube = Instance(tvtk.CubeSource, (), transient=True)
    
    cube_trans = Instance(tvtk.Transform, (), transient=True)
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('length', editor=NumEditor),
                       Item('width', editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor)
                        ),
                   )
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = [RectangularFace(owner=self, material = self.material)]
        return fl
    
    ''' #copied from cylinder, probably easy to modify for a cube
    def make_step_shape(self):
        from raypier.step_export import make_cylinder
        cyl = make_cylinder(self.centre, 
                             self.direction, 
                             self.diameter/2, 
                             self.thickness,
                             self.offset,
                             self.x_axis)
        return cyl, "green"
    '''
    @on_trait_change("length, width, offset")
    def config_pipeline(self):
        cube = self.vtk_cube
        cube.x_length = self.length
        cube.y_length = self.width
        thick = self.thickness
        cube.z_length = thick
        
        t = self.cube_trans
        t.identity()
        t.translate(self.offset,0,-thick/2) #not sure about self.offset
        t.rotate_x(180.)
        
        self.update = True
    
    def _pipeline_default(self):
        cube = self.vtk_cube
        norms = tvtk.PolyDataNormals(input_connection=cube.output_port)
        transF1 = tvtk.TransformFilter(input_connection=norms.output_port, 
                                       transform=self.cube_trans)
        transF2 = tvtk.TransformFilter(input_connection=transF1.output_port, 
                                       transform=self.transform)
        self.config_pipeline()
        return transF2
    
    def make_step_shape(self):
        """Creates an OpenCascade BRep Shape
        representation of the object, which can be
        exported to STEP format"""
        from raypier.step_export import make_box
        position = self.centre
        direction = self.direction
        x_axis = self.x_axis
        dz = self.thickness
        dx = self.length
        dy = self.width
        
        box = make_box(position, direction, x_axis, dx, dy, dz)
        
        return box, "green"
    

class RectWindow(RectMirror, Optic):
    n_inside = 1.5
    name = "Rectangular window"
    abstract = False
    
    traits_view = View(VGroup(
                       Traceable.uigroup,
                       Item('length', editor=NumEditor),
                       Item('width',editor=NumEditor),
                       Item('thickness', editor=NumEditor),
                       Item('offset', editor=NumEditor),
                       Item('n_inside', editor=NumEditor),
                        ),
                   )
                   
    vtkproperty = tvtk.Property(opacity = 0.4,
                             color = (0.8,0.8,1.0))
                             
    def _material_default(self):
        return Optic._material_default(self)
                   
    def _thickness_changed(self, new_t):
        self.faces[1].z_plane = new_t
    
    def _faces_default(self):
        m = self.material
        fl = FaceList(owner=self)
        fl.faces = [RectangularFace(owner=self, z_plane=0, material=m),
                    RectangularFace(owner=self, z_plane=-self.thickness, 
                        invert_normal=True, material=m)]
        return fl
    
            
if __name__=="__main__":
    from ray_tracer import RayTraceModel, BuildRaySet, BuildConfocalRaySet
    
    mirror = PECMirror(centre=(0,0,0),
                       orientation=0.0,
                       elevation=45.0)
    
#    input_rays = BuildRaySet(origin = (0,0,-20),
#                         direction = (0,0,1),
#                         radius=5.0,
#                         count=20)
    input_rays = BuildConfocalRaySet(focus=(0,0,5), 
                                     direction=(0,0,1),
                                     theta=20, 
                                     working_dist=20, 
                                     count=20)
    
    model = RayTraceModel(optics=[mirror], rays=input_rays)
    model.configure_traits()



from raypier.sources import ParallelRaySource
from raypier.tracer import RayTraceModel
from raypier.corner_cubes import HollowRetroreflector, SolidRetroreflector


source = ParallelRaySource(origin=(-10,0.1,50),
                           direction=(0,0,-1),
                           radius=4,
                           scale=0.1,
                           number=20)

cc = SolidRetroreflector() #
#cc = HollowRetroreflector()

data = """
- !ParallelRaySource
  __traits_version__: 3.3.0
  detail_resolution: 32
  direction: !!python/tuple [0.0, 0.0, -1.0]
  display: pipes
  export_pipes: false
  max_ray_len: 200.0
  name: Ray Source
  number: 20
  origin: !!python/tuple [-10.0, 0.10000000000000001, 50.0]
  radius: 4.0
  rings: 3
  scale: 0.10000000000000001
  scale_factor: 0.20000000000000001
  show_direction: false
  show_normals: false
  show_polarisation: false
  show_start: true
- !SolidRetroreflector
  __traits_version__: 3.3.0
  _orientation: !!python/tuple [0.0, 0.0]
  all_rays: false
  centre: !!python/tuple [0.0, 0.0, 0.0]
  diameter: 25.399999999999999
  direction: !!python/tuple [0.0, 0.0, 1.0]
  display: shaded
  elevation: 0.0
  material: !!python/object:raypier.core.cmaterials.DielectricMaterial {}
  n_inside: !!python/complex '1.5'
  n_outside: !!python/complex '1.0'
  name: Solid Retroreflector
  orientation: 0.0
  rotation: 0.0
  thickness: 20.0
"""

import yaml

print(yaml.dump([source,cc]))

source, cc = yaml.load(data)

model = RayTraceModel(sources=[source],
                      optics=[cc])

model.configure_traits()
#model.update = True
#print source.traced_rays


#    Copyright 2009, Teraview Ltd.
#
#    This file is part of Raypier.
#
#    Raypier is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from traits.api import HasTraits, Array, BaseFloat, Complex,\
            Property, List, Instance, Range, Any,\
            Tuple, Event, cached_property, Set, Int, Trait, Button,\
            self, Str, Bool, PythonValue, Enum, MetaHasTraits
from traitsui.api import View, Item, ListEditor, VSplit,\
            RangeEditor, ScrubberEditor, HSplit, VGroup, TextEditor,\
            TupleEditor, VGroup, HGroup, TreeEditor, TreeNode, TitleEditor,\
            ShellEditor
            
from traitsui.file_dialog import save_file
            
from tvtk.api import tvtk
import numpy
import threading, os, itertools
from itertools import chain, islice, count
import yaml
from abc import ABCMeta

from raypier.constraints import BaseConstraint
from raypier.has_queue import HasQueue, on_trait_change
from raypier.utils import normaliseVector, transformNormals, transformPoints,\
        transformVectors, dotprod
from raypier.core import ctracer, cmaterials
from raypier.core.ctracer import RayCollection, GaussletCollection
from .shapes import BaseShape
from raypier.vtk_algorithms import EmptyGridSource
from .editors import NumEditor, IntEditor, ComplexEditor, ROField, VectorEditor

Vector = Array(shape=(3,))

counter = count()


class BaseRayCollection(metaclass=ABCMeta):
    pass
BaseRayCollection.register(RayCollection)
BaseRayCollection.register(GaussletCollection)


class Float(BaseFloat):
    def validate(self, obj, name, value):
        return float(value)


class Direction(HasTraits):
    x = Float
    y = Float
    z = Float


class RaypierObjectMetaclass(MetaHasTraits):
    """
    The metaclass for YAMLObject.
    """
    def __init__(cls, name, bases, kwds):
        super(RaypierObjectMetaclass, cls).__init__(name, bases, kwds)
        if 'yaml_tag' in kwds and kwds['yaml_tag'] is not None:
            pass
        else:
            cls.yaml_tag = "!"+name
        cls.yaml_loader.add_constructor(cls.yaml_tag, cls.from_yaml)
        cls.yaml_dumper.add_representer(cls, cls.to_yaml)
        
        if not cls.abstract:
            cls.subclasses.add(cls)


class RaypierObject(object, metaclass=RaypierObjectMetaclass):
    """
    An object that can dump itself to a YAML stream
    and load itself from a YAML stream.
    """
    __slots__ = ()  # no direct instantiation, so allow immutable subclasses

    yaml_loader = yaml.Loader
    yaml_dumper = yaml.Dumper

    yaml_tag = None
    yaml_flow_style = None
    
    abstract = True
    subclasses = set()

    def from_yaml(cls, loader, node):
        """
        Convert a representation node to a Python object.
        """
        return loader.construct_yaml_object(node, cls)
    from_yaml = classmethod(from_yaml)

    def to_yaml(cls, dumper, data):
        """
        Convert a Python object to a representation node.
        """
        return dumper.represent_yaml_object(cls.yaml_tag, data, cls,
                flow_style=cls.yaml_flow_style)
    to_yaml = classmethod(to_yaml)
    
    @classmethod
    def class_trait_view_elements(cls):
        return None


class Renderable(HasQueue, RaypierObject, metaclass=RaypierObjectMetaclass):
    display = Enum("shaded", "wireframe", "hidden")
    
    actors = Instance(tvtk.ActorCollection, (), transient=True)
    render = Event() #request rerendering (but not necessarily re-tracing)
    
    def _display_changed(self, vnew):
        if vnew=="shaded":
            for actor in self.actors:
                actor.visibility = True
                actor.property.representation = "surface"
        elif vnew=="wireframe":
            for actor in self.actors:
                actor.visibility = True
                actor.property.representation = "wireframe"
        else:
            for actor in self.actors:
                actor.visibility = False
        self.render = True
    
    def get_actors(self, scene):
        return self.actors


class ModelObject(Renderable):
    abstract=True
    name = Str("A traceable component")
    
    update = Event() #request re-tracing
    
    centre = Tuple(0.,0.,0.) #position
    
    _orientation = Tuple(Float, Float)
    
    orientation = Property(Range(-180.0,180.0), transient=True,
                           depends_on="_orientation")
    elevation = Property(Range(-180.,180.), transient=True,
                         depends_on="_orientation")
    
    rotation = Range(-180.0,180.0, value=0.0) #rotation around orientation axis
    
    direction_btn = Button("set")
    direction = Property(Tuple(float,float,float),depends_on="_orientation")
    
    x_axis = Property(Tuple(float,float,float),depends_on="_orientation, rotation")
    
    dir_x = Property(depends_on="direction")
    dir_y = Property(depends_on="direction")
    dir_z = Property(depends_on="direction")
    
    transform = Instance(tvtk.Transform, (), transient=True)
    
    def _direction_btn_changed(self):
        d = self.direction
        D = Direction(x=d[0],y=d[1],z=d[2])
        D.edit_traits(kind='modal')
        self.direction = D.x, D.y, D.z
    
    def _get_orientation(self): return self._orientation[0]
    
    def _set_orientation(self, v): self._orientation = (v, self._orientation[1])
    
    def _get_elevation(self): return self._orientation[1]
    
    def _set_elevation(self, v): self._orientation = self._orientation[0], v
    
    def _get_dir_x(self): return self.direction[0]
    
    def _get_dir_y(self): return self.direction[1]
    
    def _get_dir_z(self): return self.direction[2]
    
    @on_trait_change("_orientation, centre, rotation")
    def on_position(self):
        trans = self.transform
        trans.identity()
        trans.translate(*self.centre)
        o,e = self._orientation
        trans.rotate_z(o)
        trans.rotate_x(e)
        trans.rotate_z(self.rotation)
        #print "set transform", self._orientation
        self.update = True
        
    def translate_local(self, du, dv, dw):
        """Translate the object along it's local axes, where
        dw is along the direction-vector, du and dv are along the orthogonal
        axes to this"""
        centre = numpy.array(self.centre)
        w = numpy.array(self.direction)
        u = numpy.array(self.x_axis)
        v = numpy.cross(w,u)
        new_centre = centre + (du*u + dv*v + dw*w)
        self.centre = tuple(new_centre)
        
    def _get_x_axis(self):
        temp = tvtk.Transform()
        o,e = self._orientation
        temp.rotate_z(o)
        temp.rotate_x(e)
        temp.rotate_z(self.rotation)
        direct = temp.transform_point(1,0,0)
        return direct
        
    @cached_property
    def _get_direction(self):
        temp = tvtk.Transform()
        o,e = self._orientation
        temp.rotate_z(o)
        temp.rotate_x(e)
        temp.rotate_z(self.rotation)
        direct = temp.transform_point(0,0,1)
        #print "get direction", direct, o, e
        return direct
    
    def _set_direction(self, d):
        x,y,z = normaliseVector(d)
        Theta = numpy.arccos(z)
        theta = 180*Theta/numpy.pi
        phi = 180*numpy.arctan2(x,y)/numpy.pi
        #print "set direction", -phi, -theta
        self._orientation = -phi, -theta
        
    def make_step_shape(self):
        """Creates an OpenCascade BRep Shape
        representation of the object, which can be
        exported to STEP format"""
        return False, None
        
        
class Probe(ModelObject):
    """An object which calculates some result from the raypier output, based on the
    probe geometry.
    
    Although probes inheret the 'update' trait from its superclass, this event doesn't trigger
    a new tracing operation.
    """ 
    
    def evaluate(self, src_list):
        """called once after all sources have been traced"""
        
    
class Traceable(ModelObject):
    vtkproperty = Instance(tvtk.Property, transient=True)
    
    intersections = List([], transient=True)
    
    material = Instance(ctracer.InterfaceMaterial)
    
    faces = Instance(ctracer.FaceList, 
                desc="Container of traceable faces (Face instances)",
                 transient = True)
    
    #all Traceables have a pipeline to generate a VTK visualisation of themselves
    pipeline = Any(transient=True) #a tvtk.DataSetAlgorithm ?
    
    polydata = Property(depends_on=['update', ])
    
    abstract=True
    subclasses = set()
    
    def _actors_default(self):
        pipeline = self.pipeline
        
        map = tvtk.PolyDataMapper(input_connection=pipeline.output_port)
        map.scalar_visibility = False
        act = tvtk.Actor(mapper=map)
        act.property = self.vtkproperty
        actors = tvtk.ActorCollection()
        actors.append(act)
        return actors
    
    def _get_polydata(self):
        self.pipeline.update()
        pd = self.pipeline.output
        #print pd
        return pd
    
    def trace_rays(self, rays):
        """traces a RayCollection.
        
        @param rays: a RayCollection instance
        @param face_id: id of the face to trace. If None, trace all faces
        
        returns - a recarray of intersetions with the same size as rays
                  dtype=(('length','d'),('face','O'),('point','d',[3,]))
                  
                  'length' is the physical length from the start of the ray
                  to the intersection
                  
                  'cell' is the ID of the intersecting face
                  
                  'point' is the position coord of the intersection (in world coords)
        """
        max_length = rays.max_length
        p1 = rays.origin
        p2 = p1 + max_length*rays.direction
        return self.intersect(p1, p2, max_length)
        
    def intersect(self, p1, p2, max_length):
        t = self.transform
        inv_t = t.linear_inverse
        P1 = transformPoints(inv_t, p1)
        P2 = transformPoints(inv_t, p2)
        
        faces = self.faces
        if len(faces)>1:
            traces = numpy.column_stack([f.intersect(P1, P2, max_length) for f in faces])   
            nearest = numpy.argmin(traces['length'], axis=1)
            ar = numpy.arange(traces.shape[0])
            shortest = traces[ar,nearest]
        else:
            shortest = faces[0].intersect(P1, P2, max_length)
            
        t_points = shortest['point']
        points = transformPoints(t, t_points)
        shortest['point'] = points
        return shortest
    
    def intersect_line(self, p1, p2):
        """Find the nearest intersection of a single line defined by two points
        p1 and p2
        
        returns a tuple (L, F, P), where L is the scalar length from p1 to the
        intersection point, F is the intersecting face and P is the intersection
        point (a 3-vector).
        """
        p1 = numpy.asarray(p1).reshape(-1,1)
        p2 = numpy.asarray(p2).reshape(-1,1)
        max_length = ((p1-p2)**2).sum(axis=0)[0]
        nearest = self.intersect(p1, p2, max_length)[0]
        return nearest['length'], nearest['cell'], nearest['point']
    
    def update_complete(self):
        pass

OrientationEditor = RangeEditor(low=-180.0, high=180.0, mode='slider')


Traceable.uigroup = VGroup(
                   Item('name', editor=TitleEditor(), springy=False,
                        show_label=False),
                   Item('display'),
                   VGroup(
                   Item('orientation', editor=OrientationEditor),
                   Item('elevation', editor=OrientationEditor),
                   Item('rotation', editor=OrientationEditor),
                   ),
                   HGroup(Item('centre', 
                               show_label=False, 
                               editor=VectorEditor,
                               springy=True), 
                          show_border=True,
                          label="Centre"),
                   HGroup(
                          VGroup(
                          Item("dir_x", style="readonly", label="x"),
                          Item("dir_y", style="readonly", label="y"),
                          Item("dir_z", style="readonly", label="z"),
                          springy=True
                          ),
                          Item('direction_btn', show_label=False, width=-60),
                          show_border=True,
                          label="Direction"
                          ),
                   label="Geometry"
                    )


class ShapedTraceable(Traceable):
    shape = Instance(BaseShape)
    
    grid_extent = Tuple((-50.,50.,-50.,50.0,-5.,5.))
    grid_resolution = Tuple((50,50,30))
    grid_in = Instance(EmptyGridSource, ())    
    
    clip = Instance(tvtk.ClipVolume, ())
    
    def eval_sag_top(self, points):
        return None
    
    def eval_sag_bottom(self, points):
        return None
    
    def eval_grid_extent(self):
        return (-50.,50.,-50.,50.0,-5.,5.)
    
    @on_trait_change("grid_extent", "grid_resolution")
    def _update_grid(self):
        grid = self.grid_in
        e = self.grid_extent
        d = self.grid_resolution
        grid.dimensions = d
        grid.spacing = (e[1]-e[0])/(d[0]-1), (e[3]-e[2])/(d[1]-1), (e[5]-e[4])/(d[2]-1)
        grid.origin = (e[0],e[2],e[4])
        grid.modified()
        self.update=True
        
    def _grid_extent_default(self):
        return self.eval_grid_extent()
    
    def build_pipeline(self, input_node):
        return input_node
    
    def _pipeline_default(self):
        grid = self.grid_in
        self._update_grid()
        
        shape = self.shape
        func = shape.impl_func
        clip = self.clip 
        clip.input_connection=grid.output_port
        clip.clip_function=func
        clip.inside_out=1
        
        attrb = tvtk.ProgrammableAttributeDataFilter(input_connection=clip.output_port)
        def update( *args ):
            in_data = attrb.get_input_data_object(0,0)
            points = in_data.points.to_array()
            top = self.eval_sag_top(points)
            bottom = -self.eval_sag_bottom(points)
            if top is None:
                top = bottom
            if bottom is None:
                bottom = top
            if top is not bottom:
                result = numpy.maximum(top, bottom) #Or maybe I need maximum
            else:
                result = top
            out = attrb.get_output_data_object(0)
            out.point_data.scalars=result
        attrb.set_execute_method(update)
        
        clip2 = tvtk.ClipDataSet(input_connection=attrb.output_port,
                                 value=0.0,
                                 inside_out=1)
        
        #topoly = tvtk.GeometryFilter(input_connection=clip2.output_port)
        topoly = tvtk.DataSetSurfaceFilter(input_connection=clip2.output_port)
        norms = tvtk.PolyDataNormals(input_connection=topoly.output_port)
        
        node = self.build_pipeline(norms)
        
        transF = tvtk.TransformFilter(input_connection=node.output_port, 
                                      transform=self.transform)
        #self.grid_extent = extent
        return transF
    
    
class Optic(Traceable):
    n_inside = Complex(1.0+0.0j) #refractive
    n_outside = Complex(1.0+0.0j)
    
    all_rays = Bool(False, desc="trace all reflected rays")
    
    vtkproperty = Instance(tvtk.Property, (), {'opacity': 0.4, 'color': (0.8,0.8,1.0)})
                             
    def _material_default(self):
        m = cmaterials.FullDielectricMaterial(n_inside = self.n_inside,
                                    n_outside = self.n_outside)
        return m
    
    def calc_refractive_index(self, wavelengths):
        """
        Evaluates an array of (complex) refractive indices.
        
        :param wavelengths: a shape=(N,1) array of wavelengths
        
        :returns: a 2-tuple representing the inside and outside
        
        refractive indices respectively. The items in the tuple can be
        either:
         
          1) an arrays with the same shape as wavelengths and with
             dtype=numpy.complex128
          2) a complex scalar
        """
        return self.n_inside, self.n_outside
    
    @on_trait_change("n_inside, n_outside")
    def n_changed(self):
        self.material.n_inside = self.n_inside
        self.material.n_outside = self.n_outside
        self.update = True
        
        
class ShapedOptic(ShapedTraceable, Optic):
    vtkproperty = Instance(tvtk.Property, (), {'opacity': 0.4, 'color': (0.8,0.8,1.0)})
    
    def _material_default(self):
        m = cmaterials.FullDielectricMaterial(n_inside = self.n_inside,
                                    n_outside = self.n_outside)
        return m
    
    def calc_refractive_index(self, wavelengths):
        """
        Evaluates an array of (complex) refractive indices.
        @param wavelengths: a shape=(N,1) array of wavelengths
        @returns: a 2-tuple representing the inside and outside
        refractive indices respectively. The items in the tuple can be
        either 1) an arrays with the same shape as wavelengths and with
                    dtype=numpy.complex128
            or 2) a complex scalar
        """
        return self.n_inside, self.n_outside
    
    @on_trait_change("n_inside, n_outside")
    def n_changed(self):
        self.material.n_inside = self.n_inside
        self.material.n_outside = self.n_outside
        self.update = True
    
    
class VTKOptic(Optic):
    """Polygonal optics using a vtkOBBTree to do the ray intersections"""
    abstract=True
    data_source = Instance(tvtk.ProgrammableSource, (), transient=True)
    
    obb = Instance(tvtk.OBBTree, (), transient=True)
    
    def __getstate__(self):
        d = super(VTKOptic, self).__getstate__()
        bad = ['polydata','pipeline','data_source','obb', 'transform']
        for b in bad:
            d.pop(b, None)
        return d
    
    def __init__(self, *args, **kwds):
        super(VTKOptic, self).__init__(*args, **kwds)
        self.on_trait_change(self.on__polydata_changed, "_polydata")
        self.on__polydata_changed()
        
    def on__polydata_changed(self):
        self.data_source.modified()
        self.update = True
    
    def _polydata_changed(self):
        obb = self.obb
        obb.free_search_structure()
        obb.data_set = self.polydata
        obb.build_locator()
        self.data_source.modified()
        
    def _pipeline_default(self):
        source = self.data_source
        def execute():
            polydata = self._polydata
            output = source.poly_data_output
            output.shallow_copy(polydata)
        source.set_execute_method(execute)
        t = self.transform
        transf = tvtk.TransformFilter(input_connection=source.output_port, 
                                      transform=t)
        tri = tvtk.TriangleFilter(input_connection=transf.output_port)
        return tri
        
    def trace_segment(self, seg, last_optic=None, last_cell=None):
        """
        Finds the intersection of the given ray-segment with the 
        object geometry data
        """
        p1 = seg.origin
        p2 = p1 + seg.MAX_RAY_LENGTH*seg.direction
        pts = tvtk.Points()
        ids = tvtk.IdList()
        ret = self.obb.intersect_with_line(p1, p2, pts, ids)
        sqrt = numpy.sqrt
        array = numpy.array
        if ret==0:
            return None
        if self is not last_optic:
            last_cell = None
        data = [ (sqrt(((array(p) - p1)**2).sum()), p, Id)
                    for p, Id in zip(pts, ids) 
                    if Id is not last_cell ]
        if not data:
            return None
        short = min(data, key=lambda a: a[0])
        return short[0], short[1], short[2], self
    
    
class Result(HasTraits, RaypierObject):
    abstract=True
    subclasses = set()
    name = Str("a result")
    
    def calc_result(self, tracer):
        """
        Called at the end of a tracing operation, so the result can 
        be evaluated
        
        @param tracer: the RayTraceModel instance
        """
        raise NotImplementedError



import vtk
from vtk.util.numpy_support import numpy_to_vtk
import numpy

x,y,z = numpy.ogrid[-10:10:50j,-10:10:50j,-10:10:50j]
data = (x**2 + y**2 + z**2)/(5**2)
data = data.ravel().astype("f")


src = vtk.vtkProgrammableSource()
###Initialise the OutputInformation on the ProgrammableSource
executive = src.GetExecutive()
outInfo = executive.GetOutputInformation(1)
outInfo.Set(executive.WHOLE_EXTENT(), 0, 49, 0, 49, 0, 49)

def make_grid():    
    sp = src.GetStructuredPointsOutput()
    sp.SetExtent(0,49,0,49,0,49)
    sp.SetOrigin(-10,-10,-10)
    sp.SetSpacing(0.4,0.4,0.4)
    sp.GetPointData().SetScalars(numpy_to_vtk(data))

src.SetExecuteMethod(make_grid)

func = vtk.vtkSphere()
func.SetCenter(0,0,0)
func.SetRadius(5.0)

clip = vtk.vtkImageMarchingCubes()
clip.SetInputConnection(0,src.GetOutputPort(1))
clip.SetValue(0,1.0)

map = vtk.vtkPolyDataMapper()
map.SetInputConnection(clip.GetOutputPort(0))
map.ScalarVisibilityOff()
 
surfaceActor = vtk.vtkActor()
surfaceActor.SetMapper(map)

ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
iren = vtk.vtkRenderWindowInteractor()
iren.SetRenderWindow(renWin)
ren.AddActor(surfaceActor)
iren.Initialize()
renWin.Render()
iren.Start()



import unittest

from raypier.tracer import RayTraceModel
from raypier.waveplates import Waveplate
from raypier.sources import ParallelRaySource, SingleRaySource
from raypier.mirrors import PECMirror

from numpy import pi
import numpy


class TestWaveplate(unittest.TestCase):
    def setUp(self):
        self.wp = Waveplate(direction=(0.,0.,1.))
        self.src = ParallelRaySource(rings=0,
                                     E_vector=(1,0,0),
                                     E1_amp = 1.0+0.0j,
                                     E2_amp = 0.0+0.0j,
                                     direction = (0,0,1),
                                     origin = (0,0,-10),
                                    )

        self.model = RayTraceModel(optics=[self.wp,],
                                   sources=[self.src])

    def test_retardance_property(self):
        wp = self.wp
        for val in [0.0, 0.25, 0.5, 0.75, 1.0]:
            wp.retardance = val
            self.assertEqual(wp.retardance, val)

    def test_input_ray(self):
        ray = self.src.input_rays[0]
        print(ray.power, ray.ellipticity, ray.E_vector, ray.E1_amp, ray.E2_amp)
        self.assertEqual(ray.ellipticity, 0.0)

    def test_zero_retardance(self):
        wp = self.wp
        wp.retardance = 0.0
        wp.rotation = 19.235
        direction=(0.,0.,1.)
        print(wp.x_axis)
        self.model.trace_all()

        ray_in = self.src.input_rays[0]
        ray_out = self.src.traced_rays[-1][0]

        ray_in.project_E(1,1,0)
        ray_out.project_E(1,1,0)

        self.assertEqual(ray_in.E_vector, ray_out.E_vector)
        self.assertAlmostEqual(ray_in.E1_amp, ray_out.E1_amp)
        self.assertAlmostEqual(ray_in.E2_amp, ray_out.E2_amp)

    def test_zero_retardance2(self):
        wp = self.wp
        wp.retardance = 0.0
        wp.rotation = 19.235
        direction=(0.,0.,-1.)
        print(wp.x_axis)
        self.model.trace_all()

        ray_in = self.src.input_rays[0]
        ray_out = self.src.traced_rays[-1][0]

        ray_in.project_E(1,1,0)
        ray_out.project_E(1,1,0)

        self.assertEqual(ray_in.E_vector, ray_out.E_vector)
        self.assertAlmostEqual(ray_in.E1_amp, ray_out.E1_amp)
        self.assertAlmostEqual(ray_in.E2_amp, ray_out.E2_amp)

    def test_half_wave_plate(self):
        wp = self.wp
        wp.retardance = 0.5
        for rot in numpy.linspace(-180,180,20):
            wp.rotation = rot
            self.model.trace_all()

            ray_in = self.src.input_rays[0]
            ray_out = self.src.traced_rays[-1][0]

            self.assertAlmostEqual(ray_in.ellipticity, ray_out.ellipticity)


class TestWaveplateBackreflection(unittest.TestCase):
    def test_quarter_retardance(self):
        src = SingleRaySource(E_vector=(1,0,0),
                                 E1_amp = 1.0+0.0j,
                                 E2_amp = 0.0+0.0j,
                                 direction = (0,0,1),
                                 origin = (0,0,-40),
                                )

        wp = Waveplate(retardance=0.25,
                       rotation = 45.0,
                       centre=(0,0,40),
                       direction=(0,0,1))

        m = PECMirror(centre=(0,0,80),
                      direction=(0.0,0.0,1.0))

        model = RayTraceModel(optics=[wp,m], sources=[src])
        #model.trace_all()

        def get_major_axis(ray):
            E1_vector = numpy.array(ray.E_vector)
            E2_vector = numpy.cross(numpy.array(ray.direction),E1_vector
                                    )
            E = (E1_vector*ray.E1_amp + E2_vector*ray.E2_amp)
            return E

        ray_in = src.input_rays[0]

        #print ray_in.E_vector, ray_in.E1_amp, ray_in.E2_amp

        self.assertEqual(len(src.traced_rays), 4)

        for i in range(4):
            ray = src.traced_rays[i][0]
            print(i, "::", ray.E_left, ray.E_right, ray.E_vector, ray.E1_amp, ray.E2_amp)
            print("\t", get_major_axis(ray))

        ray_out = src.traced_rays[-1][0]
        self.assertAlmostEqual(numpy.dot(get_major_axis(ray_out),
                                    get_major_axis(ray_in)), 0.0)

    def test_quarter_retardance2(self):
        src = SingleRaySource(E_vector=(1,0,0),
                                 E1_amp = 1.0+0.0j,
                                 E2_amp = 0.0+0.0j,
                                 direction = (0,0,1),
                                 origin = (0,0,-40),
                                )

        wp1 = Waveplate(retardance=0.25,
                       rotation = 45.0,
                       centre=(0,0,40),
                       direction=(0,0,1))

        print("wp1 fa:", wp1.material.fast_axis)

        wp2 = Waveplate(retardance=0.25,
                       rotation = -45.0,
                       centre=(0,0,80),
                       direction=(0,0,-1))

        print("wp2 fa:", wp2.material.fast_axis)
        print("wp1 fa:", wp1.material.fast_axis)

        model = RayTraceModel(optics=[wp1,wp2], sources=[src])
        model.trace_all()

        def get_major_axis(ray):
            E1_vector = numpy.array(ray.E_vector)
            E2_vector = numpy.cross(numpy.array(ray.direction),
                                    E1_vector)
            E = (E1_vector*ray.E1_amp + E2_vector*ray.E2_amp)
            return E.real

        ray_in = src.input_rays[0]

        #print ray_in.E_vector, ray_in.E1_amp, ray_in.E2_amp

        self.assertEqual(len(src.traced_rays), 3)

        for i in range(len(src.traced_rays)):
            ray = src.traced_rays[i][0]
            print(i, "::", get_major_axis(ray), ray.E_left, ray.E_right, ray.ellipticity, ray.E_vector)

if __name__ == "__main__":
    unittest.main()





from traits.api import Float, Str, Instance
from traitsui.api import Item, View, VGroup

from tvtk.api import tvtk

from .core.cfaces import CircularFace
from .core.cmaterials import CoatedDispersiveMaterial
from .dispersion import NamedDispersionCurve, NondispersiveCurve
from .bases import Optic, Traceable, NumEditor
from .core.ctracer import FaceList


class CircularWindow(Optic):
    abstract = False
    
    thickness = Float(5.0, desc="centre thickness")
    diameter = Float(15.0)
    offset = Float(0.0)
    
    glass_type = Str("N-BK7")
    coating_thickness = Float(0.283) #microns
    coating_refractive_index = Float(1.37)
    
    vtk_cylinder = Instance(tvtk.Cylinder, ())
    
    traits_view = View(VGroup(
                   Traceable.uigroup,  
                   Item('glass_type'),
                   Item('thickness', editor=NumEditor),
                   Item('diameter', editor=NumEditor),
                   Item('offset', editor=NumEditor)
                   )
                )
    

    def _material_default(self):
        glass = NamedDispersionCurve(self.glass_type)
        cri = self.coating_refractive_index
        coating = NondispersiveCurve(refractive_index=cri)
        air = NondispersiveCurve(1.0)
        
        m = CoatedDispersiveMaterial(dispersion_inside=glass,
                                     dispersion_outside=air,
                                     dispersion_coating=coating)
        return m
    
    def _faces_default(self):
        fl = FaceList(owner=self)
        fl.faces = [CircularFace(owner=self, diameter=self.diameter,
                                 offset = self.offset, z_plane=0.0,
                                material = self.material), 
                    CircularFace(owner=self, diameter=self.diameter,
                                material=self.material, offset=self.offset,
                                z_plane=self.thickness)]
        return fl
    
    




import unittest

from raypier.sources import HexagonalRayFieldSource
from raypier.fields import evaluate_neighbours, project_to_sphere,\
        evaluate_modes
from raypier.core.cfields import evaluate_modes as evaluate_modes_c
        
#from raypier.core.cfields import inv_area_of_ellipse

import numpy


class TestProjectToSphere(unittest.TestCase):
    def test_project1(self):
        src = HexagonalRayFieldSource(centre=(0,0,-50),
                                      direction=(0,0,1),
                                      spacing=2,
                                      radius=10.0)
        
        rays = src.input_rays.copy_as_array()
        phase = rays['phase'] #will be zeros
        wavelengths= src.wavelength_list
        
        out_rays, out_phase = project_to_sphere(rays, phase, wavelengths, 
                                     centre=(0,0,0), 
                                     radius=20.0)
        
        print(out_rays['origin'])
        self.assertAlmostEqual(out_rays['origin'][:,2].min(), -20.0)
        
        
class TestEvalNeighbours(unittest.TestCase):
    def test_eval_neighbours(self):
        src = HexagonalRayFieldSource(centre=(0,0,-50),
                                      direction=(0,0,1),
                                      spacing=4,
                                      radius=10.0)
        
        rays_in = src.input_rays.copy_as_array()
        nb_idx = src.neighbours

        self.assertEquals(len(rays_in), 19)
        rays, x, y, dx, dy = evaluate_neighbours(rays_in, nb_idx)
        self.assertEquals(len(rays), 7)
        self.assertEqual( len(x.shape), 2)
        
        offset = x**2 + y**2
        self.assertTrue( numpy.allclose(offset,offset.mean()) )
        
        self.assertTrue( numpy.allclose(dx, 0) )
        self.assertTrue( numpy.allclose(dy, 0) )
        
        
class TestEvalModes(unittest.TestCase):
    def test_solve_modes(self):
        src = HexagonalRayFieldSource(centre=(0,0,-50),
                                      direction=(0,0,1),
                                      spacing=4,
                                      radius=10.0)
        
        rays_in = src.input_rays.copy_as_array()
        nb_idx = src.neighbours

        self.assertEquals(len(rays_in), 19)
        rays, x, y, dx, dy = evaluate_neighbours(rays_in, nb_idx)
        print(x.shape)
        Z = evaluate_modes(rays, x, y, dx, dy)
        print(Z.shape)
        
    def test_eval_modes_c(self):
        angles = numpy.array([numpy.linspace(0,numpy.pi*2, 7)[:6]])
        x = numpy.sin(angles)*2.0
        y = numpy.cos(angles)
        rot = 20.0*numpy.pi/180.0
        x2 = numpy.cos(rot)*x -numpy.sin(rot)*y
        y2 = numpy.cos(rot)*y + numpy.sin(rot)*x
        dx = -x/10.0 + y/50.0
        dy = y/10.0 + x/50.0
        
        ret1 = evaluate_modes(x2, y2, dx, dy)
        print(ret1)
        ret2 = evaluate_modes_c(x2,y2,dx,dy)
        print(ret2)
        self.assertTrue(numpy.allclose(ret1, ret2))
        
    def test_eval_modes_c_scale(self):
        
        xyz = numpy.array([[[ 0.        ,  0.1       ,  0.        ],
                    [ 0.        ,  0.1       ,  0.        ],
                    [-0.08660254, -0.05      ,  0.        ],
                    [-0.08660254, -0.05      ,  0.        ],
                    [ 0.08660254, -0.05      ,  0.        ],
                    [ 0.08660254, -0.05      ,  0.        ]]])
        print(xyz.shape)
        x=xyz[:,:,0]
        y=xyz[:,:,1]
        dx =dy = numpy.zeros_like(x)
        
        ret1 = evaluate_modes_c(x, y, dx, dy)
        print(ret1)
        
        xyz = numpy.array([[[-8.27605704e-02,  1.00000000e-01,  2.60000000e+01],
                        [ 8.27605704e-02,  1.00000000e-01,  2.60000000e+01],
                        [-4.52222552e-02,  2.16727564e-02,  2.60000000e+01],
                        [-1.27982826e-01, -1.21672756e-01,  2.60000000e+01],
                        [ 1.27982826e-01, -1.21672756e-01,  2.60000000e+01],
                        [ 4.52222552e-02,  2.16727564e-02,  2.60000000e+01]]])
        x=xyz[:,:,0]
        y=xyz[:,:,1]
        dx =dy = numpy.zeros_like(x)
        
        ret1 = evaluate_modes_c(x, y, dx, dy)
        print(ret1)
        
        
class TestAreaOfEllipse(unittest.TestCase):
    def test_area_of_ellipse(self):
        ra = 3.0
        rb = 8.0
        angle = 30*numpy.pi/180.0
        
        a=ra #(1/(ra**2))
        b=rb #(1/(rb**2))
        
        sin = numpy.sin(angle)
        cos = numpy.cos(angle)
        
        A = (a**2)*(sin**2) + (b**2)*(cos**2)
        B = 2*(b**2 - a**2)*sin*cos
        C = (a**2)*(cos**2) + (b**2)*(sin**2)
        
        area = numpy.pi*ra*rb
        print("True area:", area)
        
        area2 = (inv_area_of_ellipse(A, B, C)**1)
        
        self.assertAlmostEqual(area, area2)
